from __future__ import division, print_function

from datetime import datetime, timedelta
from typing import Annotated, Dict, List, Literal, Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly.graph_objs as go
import seaborn as sns
from plotly.subplots import make_subplots
from scipy.spatial import ConvexHull
from sklearn.cluster import KMeans

sns.set_style("whitegrid")

import matplotlib.colors as mcolors
from scipy import signal as sig
from scipy.interpolate import CubicSpline
from scipy.optimize import fmin
from sklearn.decomposition import PCA
from sklearn.linear_model import LinearRegression
from sklearn.metrics import mean_squared_error
from sklearn.preprocessing import StandardScaler

from CurvyCUSIPs.utils.ust_utils import enhanced_plotly_blue_scale


# TODO
def covar_matrix_estimates():
    pass


def calc_pca_loadings_matrix(
    df: pd.DataFrame,
    cols: Optional[List[str]] = None,
    run_on_level_changes: Optional[bool] = False,
    run_pca_on_corr_mat: Optional[bool] = False,
    scale_loadings: Optional[bool] = False,
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    if cols:
        df = df[cols].copy()
    else:
        df = df.copy()

    if run_on_level_changes:
        df = df.diff().dropna()

    if run_pca_on_corr_mat:
        scaler = StandardScaler()
        values_to_fit = scaler.fit_transform(df)
    else:
        values_to_fit = df.values

    pca = PCA()
    pc_scores_signs_not_flipped = pca.fit_transform(values_to_fit)
    pc_scores_df = pd.DataFrame(
        {
            "Date": df.index,
            "PC1": pc_scores_signs_not_flipped[:, 0],
            # https://stackoverflow.com/questions/44765682/in-sklearn-decomposition-pca-why-are-components-negative
            "PC2": -1 * pc_scores_signs_not_flipped[:, 1],
            "PC3": -1 * pc_scores_signs_not_flipped[:, 2] if len(df.columns) > 3 else None,
        }
    )

    if scale_loadings:
        # the sensitivity of each original tenor to changes in each principal component
        # the loadings reflect how much variance (in original units) is explained by each PC b/c scaled by sqrt(eigenvalues)
        loadings_df = pd.DataFrame(
            pca.components_.T * np.sqrt(pca.explained_variance_), index=df.columns, columns=[f"PC{i+1}" for i in range(len(df.columns))]
        )
    else:
        loadings_df = pd.DataFrame(pca.components_.T, index=df.columns, columns=[f"PC{i+1}" for i in range(len(df.columns))])

    return {
        "pca": pca,
        "loading_matrix": loadings_df,
        "pc_scores_matrix": pc_scores_df,
        "covar_matrix": pd.DataFrame(pca.get_covariance(), index=df.columns, columns=df.columns),
    }


def split_dates_into_ranges(dates: List[datetime]) -> List[List[datetime]]:
    if not dates:
        return []

    sorted_dates = sorted(dates)
    ranges = []
    current_range = [sorted_dates[0]]

    for i in range(1, len(sorted_dates)):
        if sorted_dates[i] - sorted_dates[i - 1] == timedelta(days=1):
            current_range.append(sorted_dates[i])
        else:
            ranges.append(current_range)
            current_range = [sorted_dates[i]]

    if current_range:
        ranges.append(current_range)

    return ranges


def identify_yield_curve_movements(
    pc1: pd.Series,
    pc2: pd.Series,
    date_index: pd.DatetimeIndex,
    resample: Optional[Literal["W", "M", "Y"]] = None,
) -> Annotated[List[List[datetime]], 4]:
    """
    0 => Bull Steepening => lime
    1 => Bear Steepening => skyblue
    2 => Bull Flattening => coral
    4 => Beat Flattening => khaki
    """
    movements = [[], [], [], []]

    if resample:
        df = pd.DataFrame({"PC1": pc1, "PC2": pc2}, index=date_index)
        resample_df = df.resample(resample).last()
        resampled_changes = resample_df.diff()
        for i in range(1, len(resampled_changes)):
            if resampled_changes["PC1"].iloc[i] < 0 and resampled_changes["PC2"].iloc[i] > 0:
                movements[0].append(resampled_changes.index[i])
            elif resampled_changes["PC1"].iloc[i] > 0 and resampled_changes["PC2"].iloc[i] > 0:
                movements[1].append(resampled_changes.index[i])
            elif resampled_changes["PC1"].iloc[i] < 0 and resampled_changes["PC2"].iloc[i] < 0:
                movements[2].append(resampled_changes.index[i])
            elif resampled_changes["PC1"].iloc[i] > 0 and resampled_changes["PC2"].iloc[i] < 0:
                movements[3].append(resampled_changes.index[i])
    else:
        for i in range(1, len(pc1)):
            if pc1[i] < pc1[i - 1] and pc2[i] > pc2[i - 1]:
                movements[0].append(date_index[i])
            elif pc1[i] > pc1[i - 1] and pc2[i] > pc2[i - 1]:
                movements[1].append(date_index[i])
            elif pc1[i] < pc1[i - 1] and pc2[i] < pc2[i - 1]:
                movements[2].append(date_index[i])
            elif pc1[i] > pc1[i - 1] and pc2[i] < pc2[i - 1]:
                movements[3].append(date_index[i])

    return movements


"""
Salomon PCA Note:

    - by def, pca does not recognize time series aspect of term structure data (can scramble and still get same PCs)
    - implicit assumption in PCA is that the data are independent and identically distributed
        - violated if vol is depended on level
        - vol/changes depending on levels is the gist of mean-reverting termstructure models
        - independent and identically distributed changes is consistent with a random walk model (negative yields allowed)
        - simple-minded PCA on yield changes cannot be consistent with a meanreverting framework
            - vol/changes would exhibit positive correlation/trend if levels were either at too high or too a low value in their ranges
        - depends of current macro env => curve shape
        - in typical/normal env,  levels are more likely to take values around this typical shape than around extreme curve shapes
        - change data will likely be dominated by samples with little or no observable serial correlation
        - above two points provides some comfort about the assumption of independence in the change data
        
    - eigenvalue stability decays over time and at higher orders
    - interpretation of the exposure to the fourth PC is provided by viewing the exposure as relative value in a portfolio

    for flies:
    - use pc on levels for flies vs changes (to generate signals and compute beta weights)
    - comparsion with regression beta weights
    - inherent mean reversion assumption
    - pca used for richness and cheapness screener (see Huggins and Schaller)
    - pca used for beta weight calc

    - salomon butterfly model graphs (ex 4s10s27s)
        - long spread vs short spread (10s27s vs 4s10s) and residual
        - butterfly (dd weighted (50-50), diff beta weighted)
        - yield and slope (slope = long wing - short wing) and belly
        - butterfly (diff weightings) vs yield and residual 
        - butterfly (diff weightings) vs curve slope and residual 


CS PCA Note:


"""
# BIG TODO break this function up!


def run_pca_yield_curve(
    df: pd.DataFrame,
    date_subset_range: Annotated[List[datetime], 2] | None = None,
    n_components: Optional[int] = 3,
    run_on_diff: Optional[bool] = False,  # flag to use time step wise changes for PCA
    run_on_scaler: Optional[bool] = False,  # flag to use corr matrix for PCA
    show_cumul_ex_var: Optional[bool] = False,
    show_eigenvectors: Optional[bool] = False,
    show_hedge_ratios: Optional[bool] = False,
    show_3d_plot: Optional[bool] = False,
    show_pc_scores_timeseries: Optional[bool] = False,
    show_recessions: Optional[bool] = False,
    window: Optional[int] = None,
    show_most_recent: Optional[bool] = False,
    show_trend: Optional[bool] = False,
    show_reconstructed: Optional[bool] = False,
    curve_analysis_resampling_window: Optional[Literal["W", "M", "Y"]] = None,
    show_bull_steepening_periods: Optional[bool] = False,
    # show_bear_steepening_periods: Optional[bool] = False,
    # show_bull_flattening_periods: Optional[bool] = False,
    # show_bear_flattening_periods: Optional[bool] = False,
    is_cmt_df=False,
    is_cusips=False,
    html_path: Optional[str] = None,
    show_clusters: Optional[bool] = False,
    num_clusters=8,
    overlay_df: Optional[pd.DataFrame] = None,
    to_overlay_pcs_v_time_cols: Optional[Annotated[List[str], 3]] = None,
):
    df = df.copy()
    if date_subset_range:
        df = df.loc[date_subset_range[0] : date_subset_range[1]]

    to_return_dict: Dict[str, pd.DataFrame | List] = {
        "reconstructed_df": None,
        "factor_loadings_df": None,
        "biplot_df": None,
        "cumulative_explained_variance": None,
        "explained_variance": None,
    }

    if run_on_diff:
        df = df.diff()
        df = df.dropna()

    if run_on_scaler:
        scaler = StandardScaler()
        data_scaled = scaler.fit_transform(df)
    else:
        data_scaled = df.values

    if show_cumul_ex_var:
        pca = PCA()
        pca.fit(data_scaled)
        cumulative_explained_variance = np.cumsum(pca.explained_variance_ratio_)
        explained_variance = pca.explained_variance_ratio_
        to_return_dict["cumulative_explained_variance"] = cumulative_explained_variance
        to_return_dict["explained_variance"] = explained_variance

        _, (ax1, ax2) = plt.subplots(2, 1, figsize=(10, 10))
        ax1.plot(cumulative_explained_variance, marker="o")
        ax1.set_xlabel("Number of Components")
        ax1.set_ylabel("Cumulative Explained Variance")
        ax1.set_title("Cumulative Explained Variance by PCA Components")
        ax1.grid(True)
        ax1.set_xticks(range(1, len(explained_variance) + 1))

        ax2.bar(range(1, len(explained_variance) + 1), explained_variance, alpha=0.7)
        ax2.set_xlabel("Number of Components")
        ax2.set_ylabel("Individual Explained Variance")
        ax2.set_title("Individual Explained Variance by PCA Components")
        ax2.grid(True)
        ax2.set_xticks(range(1, len(explained_variance) + 1))

        plt.tight_layout()
        plt.show()

    pca_obj = PCA(n_components=n_components)
    data_fitted_transf_pca = pca_obj.fit_transform(data_scaled)

    if show_eigenvectors:
        tenors_label = df.columns
        fig, axes = plt.subplots(1, 3, figsize=(17, 10))
        pclabels = ["Level", "Slope", "Curvature"]
        for i in range(3):
            axes[i].plot(tenors_label, pca_obj.components_[i, :])
            axes[i].set_title(f"Principal Component {i+1} ({pclabels[i]})")
            axes[i].set_xlabel("Tenor")
            axes[i].set_ylabel("Loading (bps)" if not run_on_diff else "Loading (bps change)")
            axes[i].set_xticks(tenors_label)
            axes[i].tick_params(axis="x", rotation=45)
            axes[i].grid(True)

        plt.tight_layout()
        plt.show()

        factor_loadings_df = pd.DataFrame(pca_obj.components_, columns=tenors_label)
        factor_loadings_df.index = [f"PC{i+1}" for i in range(pca_obj.n_components_)]
        to_return_dict["factor_loadings_df"] = factor_loadings_df

    if show_reconstructed:
        reconstructed_values = pca_obj.inverse_transform(data_fitted_transf_pca)
        reconstructed_df = pd.DataFrame(reconstructed_values, columns=df.columns, index=df.index)
        to_return_dict["reconstructed_df"] = reconstructed_df
        reconstruction_error = mean_squared_error(df, reconstructed_values)
        print("\nReconstruction Error (MSE):", reconstruction_error)

    if show_hedge_ratios:
        pass

    # biplot
    if show_3d_plot:
        biplot_df = pd.DataFrame(
            {
                "Date": df.index,
                "PC1": data_fitted_transf_pca[:, 0],
                # flipping signs here
                # https://stackoverflow.com/questions/44765682/in-sklearn-decomposition-pca-why-are-components-negative
                "PC2": -1 * data_fitted_transf_pca[:, 1],
                "PC3": -1 * data_fitted_transf_pca[:, 2],
            }
        )

        if is_cmt_df:
            default_tenor_vect_color = {
                "CMT3M": "red",
                "CMT6M": "red",
                "CMT1": "red",
                "CMT2": "orange",
                "CMT3": "orange",
                "CMT5": "orange",
                "CMT7": "orange",
                "CMT10": "cornflowerblue",
                "CMT20": "cornflowerblue",
                "CMT30": "cornflowerblue",
            }
        else:
            default_tenor_vect_color = {
                "CT3M": "red",
                "CT6M": "red",
                "CT1": "red",
                "CT2": "red",
                "CT3": "orange",
                "CT5": "orange",
                "CT7": "orange",
                "CT10": "cornflowerblue",
                "CT20": "cornflowerblue",
                "CT30": "cornflowerblue",
            }

        if show_clusters:
            kmeans = KMeans(n_clusters=num_clusters, random_state=0, n_init="auto").fit(biplot_df.set_index("Date"))
            biplot_df["cluster"] = kmeans.labels_
            to_return_dict["biplot_df"] = biplot_df.copy()
            hover_text = [
                f"Date: {row["Date"]}<br>"
                + f"PC1: {row["PC1"]:.4f}<br>"
                + f"PC2: {row["PC2"]:.4f}<br>"
                + f"PC3: {row["PC3"]:.4f}<br>"
                + f"Cluster: {row["cluster"]}"
                for _, row in biplot_df.iterrows()
            ]
            biplot_df["hover_text"] = hover_text
        else:
            hover_text = [
                f"Date: {row["Date"]}<br>" + f"PC1: {row["PC1"]:.4f}<br>" + f"PC2: {row["PC2"]:.4f}<br>" + f"PC3: {row["PC3"]:.4f}"
                for _, row in biplot_df.iterrows()
            ]
            to_return_dict["biplot_df"] = biplot_df.copy()

        traces = []
        if show_clusters:
            biplot_df = biplot_df.sort_values(by=["cluster"])
            for cluster_id in biplot_df["cluster"].unique():
                cluster_data = biplot_df[biplot_df["cluster"] == cluster_id]
                trace = go.Scatter3d(
                    x=cluster_data["PC1"],
                    y=cluster_data["PC2"],
                    z=cluster_data["PC3"],
                    mode="markers",
                    marker=dict(size=4, color=cluster_id, colorscale="RdBu", opacity=0.8),
                    text=cluster_data["hover_text"],
                    hoverinfo="text",
                    name=f"Cluster {cluster_id}",
                )
                traces.append(trace)
        else:
            scatter = go.Scatter3d(
                x=biplot_df["PC1"],
                y=biplot_df["PC2"],
                z=biplot_df["PC3"],
                mode="markers",
                marker=dict(
                    size=4,
                    color=biplot_df["PC2"],
                    colorscale="RdBu",
                    opacity=0.8,
                ),
                text=hover_text,
                hoverinfo="text",
            )
            traces.append(scatter)

        vectors = []
        for i, tenor in enumerate(df.columns):
            # sign flipping here maybe?
            vectors.append(
                go.Scatter3d(
                    x=[0, pca_obj.components_[0, i]],
                    y=[0, pca_obj.components_[1, i]],
                    z=[0, pca_obj.components_[2, i]],
                    mode="lines+text",
                    line=dict(color=default_tenor_vect_color[tenor], width=3) if not is_cusips else None,
                    text=["", tenor],
                    textposition="top center",
                    textfont=dict(color=default_tenor_vect_color[tenor]) if not is_cusips else None,
                    hoverinfo="text",
                    hoverlabel=dict(bgcolor=default_tenor_vect_color[tenor]) if not is_cusips else None,
                    name=tenor,
                )
            )

        layout = go.Layout(
            scene=dict(
                xaxis_title=f"PC1 ({pca_obj.explained_variance_ratio_[0]:.2%})",
                yaxis_title=f"PC2 ({pca_obj.explained_variance_ratio_[1]:.2%})",
                zaxis_title=f"PC3 ({pca_obj.explained_variance_ratio_[2]:.2%})",
                aspectmode="manual",
                aspectratio=dict(x=2, y=2, z=1),
            ),
            title="Interactive 3D PCA Biplot of CT Yields",
            height=1000,
            # width=1400,
            template="plotly_dark",
        )
        data = traces + vectors
        fig = go.Figure(data=data, layout=layout)
        fig.update_xaxes(showspikes=True, spikecolor="white", spikesnap="cursor", spikemode="across")
        fig.update_yaxes(showspikes=True, spikecolor="white", spikesnap="cursor", spikethickness=0.5)
        if html_path:
            fig.write_html(html_path)
        fig.show()

    if show_pc_scores_timeseries:
        pc1 = data_fitted_transf_pca[:, 0]
        # flipping signs
        pc2 = -1 * data_fitted_transf_pca[:, 1]
        pc3 = -1 * data_fitted_transf_pca[:, 2]

        if window:
            moving_avg_pc1 = pd.Series(pc1).rolling(window=window).mean()
            moving_avg_pc2 = pd.Series(pc2).rolling(window=window).mean()
            moving_avg_pc3 = pd.Series(pc3).rolling(window=window).mean()

        time_values = np.arange(len(df.index)).reshape(-1, 1)

        model_pc1 = LinearRegression()
        model_pc2 = LinearRegression()
        model_pc3 = LinearRegression()
        model_pc1.fit(time_values, pc1)
        model_pc2.fit(time_values, pc2)
        model_pc3.fit(time_values, pc3)
        trend_pc1 = model_pc1.predict(time_values)
        trend_pc2 = model_pc2.predict(time_values)
        trend_pc3 = model_pc3.predict(time_values)

        _, axes = plt.subplots(3, 1, figsize=(15, 20))

        pca_container = [
            {
                "label": "PC1 - Level",
                "pc": pc1,
                "model": model_pc1,
                "trend": trend_pc1,
                "ma": moving_avg_pc1 if window else None,
                "color": "blue",
            },
            {
                "label": "PC2 - Slope",
                "pc": pc2,
                "model": model_pc2,
                "trend": trend_pc2,
                "ma": moving_avg_pc2 if window else None,
                "color": "green",
            },
            {
                "label": "PC3 - Curvature",
                "pc": pc3,
                "model": model_pc3,
                "trend": trend_pc3,
                "ma": moving_avg_pc3 if window else None,
                "color": "red",
            },
        ]

        opp_colors = {
            "blue": "blueviolet",
            "green": "cyan",
            "red": "orange",
        }

        for i in [0, 1, 2]:
            pca_container[i]["label"]
            axes[i].plot(
                df.index,
                pca_container[i]["pc"],
                color=pca_container[i]["color"],
                label=pca_container[i]["label"],
            )
            if to_overlay_pcs_v_time_cols and overlay_df is not None:
                ax2 = axes[i].twinx()
                ax2.plot(
                    overlay_df.index,
                    overlay_df[to_overlay_pcs_v_time_cols[i]],
                    color=opp_colors[pca_container[i]["color"]],
                    label=to_overlay_pcs_v_time_cols[i],
                )
                ax2.set_ylabel(to_overlay_pcs_v_time_cols[i], color=opp_colors[pca_container[i]["color"]])

            recent = pca_container[i]["trend"][-1]
            if show_most_recent:
                axes[i].axhline(
                    pca_container[i]["pc"][0],
                    color="grey",
                    linestyle="--",
                    label=f"Recent ({recent:.2f})",
                )
            if show_trend:
                axes[i].plot(
                    df.index,
                    pca_container[i]["trend"],
                    color="black",
                    linestyle="--",
                    label="Trend",
                )
            if window:
                rolling_ma = round(pca_container[i]["ma"].tail(1).item(), 3)
                axes[i].plot(
                    df.index,
                    pca_container[i]["ma"].shift(-(window - 1)),
                    color="plum",
                    linestyle="--",
                    label=f"{window} Day Moving Average: {rolling_ma}",
                )

            if show_recessions:
                recessions_list = [
                    [datetime(1990, 7, 1), datetime(1991, 3, 1)],
                    [datetime(2001, 3, 1), datetime(2001, 11, 1)],
                    [datetime(2007, 12, 1), datetime(2009, 6, 1)],
                    [datetime(2020, 2, 1), datetime(2020, 4, 1)],
                ]
                if date_subset_range:
                    start_plot_range, end_plot_range = min(date_subset_range), max(date_subset_range)
                else:
                    start_plot_range, end_plot_range = min(df.index), max(df.index)

                for start, end in recessions_list:
                    if start <= end_plot_range and end >= start_plot_range:
                        axes[i].axvspan(start, end, color="lightcoral", alpha=0.3)

            axes[i].set_title(f"PC{i+1} over Time")
            axes[i].set_ylabel(f"PC{i+1} Score", color=pca_container[i]["color"])

            if to_overlay_pcs_v_time_cols and overlay_df is not None:
                lines, labels = axes[i].get_legend_handles_labels()
                lines2, labels2 = ax2.get_legend_handles_labels()
                ax2.legend(lines + lines2, labels + labels2, loc=0)
            else:
                axes[i].legend()

            axes[i].grid(True)

        plt.tight_layout()
        plt.show()

    return to_return_dict


def create_residuals_surface_plot(residuals_df, tick_freq=30):
    fig = go.Figure(
        data=[
            go.Surface(
                z=residuals_df.values,
                x=residuals_df.columns,
                y=residuals_df.index,
                colorscale=enhanced_plotly_blue_scale(),
                cmin=-np.max(np.abs(residuals_df.values)),
                cmax=np.max(np.abs(residuals_df.values)),
                customdata=residuals_df[["CT1"]],
                hovertemplate="Date: %{y}" + "<br>Tenor: %{x}" + "<br>Residual: %{z}<extra></extra>",
            )
        ]
    )
    fig.update_traces(
        contours_z=dict(show=True, usecolormap=True, highlightcolor="limegreen", project_z=True),
    )

    fig.update_layout(
        title="UST Yield Curve Residuals - 3D Surface Plot",
        autosize=False,
        width=1400,
        height=1000,
        scene=dict(
            xaxis_title="Tenor",
            xaxis=dict(
                tickmode="array",
                tickvals=residuals_df.columns,
            ),
            yaxis_title="Date",
            zaxis_title="Residual Value",
            camera=dict(eye=dict(x=1.6, y=-1.6, z=0.7)),
            aspectmode="manual",
            aspectratio=dict(x=2, y=2, z=1),
        ),
        template="plotly_dark",
    )
    fig.update_traces(colorbar=dict(title="Residual Value", thickness=20, len=0.75, x=0.95))
    fig.update_xaxes(showspikes=True, spikecolor="white", spikesnap="cursor", spikemode="across")
    fig.update_yaxes(showspikes=True, spikecolor="white", spikesnap="cursor", spikethickness=0.5)
    return fig


def plot_residuals_for_date(residuals_df, date):
    if date not in residuals_df.index:
        raise ValueError(f"Date {date} not found in the residuals DataFrame.")

    residuals_for_date = residuals_df.loc[date]
    fig = go.Figure(
        data=[
            go.Bar(
                x=residuals_df.columns,
                y=residuals_for_date,
                marker=dict(
                    color=residuals_for_date,
                    colorscale="RdBu",
                    colorbar=dict(title="Residual Value (bps)"),
                ),
            )
        ]
    )
    fig.update_layout(
        title=f"Residuals for {date}",
        xaxis_title="Tenor",
        yaxis_title="Residual Value",
        template="plotly_dark",
    )
    return fig



"""
Base class for cointegration approach in statistical arbitrage.
"""

from abc import ABC
import pandas as pd


def construct_spread(price_data: pd.DataFrame, hedge_ratios: pd.Series, dependent_variable: str = None) -> pd.Series:
    """
    Construct spread from `price_data` and `hedge_ratios`. If a user sets `dependent_variable` it means that a
    spread will be:

    hedge_ratio_dependent_variable * dependent_variable - sum(hedge_ratios * other variables).
    Otherwise, spread is:  hedge_ratio_0 * variable_0 - sum(hedge ratios * variables[1:]).

    :param price_data: (pd.DataFrame) Asset prices data frame.
    :param hedge_ratios: (pd.Series) Hedge ratios series (index-tickers, values-hedge ratios).
    :param dependent_variable: (str) Dependent variable to use. Set None for dependent variable being equal to 0 column.
    :return: (pd.Series) Spread series.
    """

    weighted_prices = price_data * hedge_ratios  # price * hedge

    if dependent_variable is not None:
        non_dependent_variables = [x for x in weighted_prices.columns if x != dependent_variable]
        return weighted_prices[dependent_variable] - weighted_prices[non_dependent_variables].sum(axis=1)

    return weighted_prices[weighted_prices.columns[0]] - weighted_prices[weighted_prices.columns[1:]].sum(axis=1)


class CointegratedPortfolio(ABC):
    """
    Class for portfolios formed using the cointegration method (Johansen test, Engle-Granger test).
    """

    def construct_mean_reverting_portfolio(self, price_data: pd.DataFrame, cointegration_vector: pd.Series = None) -> pd.Series:
        """
        When cointegration vector was formed, this function is used to multiply asset prices by cointegration vector
        to form mean-reverting portfolio which is analyzed for possible trade signals.

        :param price_data: (pd.DataFrame) Price data with columns containing asset prices.
        :param cointegration_vector: (pd.Series) Cointegration vector used to form a mean-reverting portfolio.
            If None, a cointegration vector with maximum eigenvalue from fit() method is used.
        :return: (pd.Series) Cointegrated portfolio dollar value.
        """

        if cointegration_vector is None:
            cointegration_vector = self.cointegration_vectors.iloc[0]  # Use eigenvector with biggest eigenvalue.

        return (cointegration_vector * price_data).sum(axis=1)

    def get_scaled_cointegration_vector(self, cointegration_vector: pd.Series = None) -> pd.Series:
        """
        This function returns the scaled values of the cointegration vector in terms of how many units of other
        cointegrated assets should be bought if we buy one unit of one asset.

        :param cointegration_vector: (pd.Series) Cointegration vector used to form a mean-reverting portfolio.
            If None, a cointegration vector with maximum eigenvalue from fit() method is used.
        :return: (pd.Series) The scaled cointegration vector values.
        """

        if cointegration_vector is None:
            cointegration_vector = self.cointegration_vectors.iloc[0]  # Use eigenvector with biggest eigenvalue

        scaling_coefficient = 1 / cointegration_vector.iloc[0]  # Calculating the scaling coefficient

        # Calculating the scaled cointegration vector
        scaled_cointegration_vector = cointegration_vector * scaling_coefficient

        return scaled_cointegration_vector


import pandas as pd
from statsmodels.tsa.vector_ar.vecm import coint_johansen


class JohansenPortfolio(CointegratedPortfolio):
    """
    The class implements the construction of a mean-reverting portfolio using eigenvectors from
    the Johansen cointegration test. It also checks Johansen (eigenvalue and trace statistic) tests
    for the presence of cointegration for a given set of assets.
    """

    def __init__(self):
        """
        Class constructor.
        """

        self.price_data = None  # pd.DataFrame with price data used to fit the model.
        self.cointegration_vectors = None  # Johansen eigenvectors used to form mean-reverting portfolios.
        self.hedge_ratios = None  # Johansen hedge ratios.
        self.johansen_trace_statistic = None  # Trace statistic data frame for each asset used to test for cointegration.
        self.johansen_eigen_statistic = None  # Eigenvalue statistic data frame for each asset used to test for cointeg.

    def fit(self, price_data: pd.DataFrame, dependent_variable: str = None, det_order: int = 0, n_lags: int = 1):
        """
        Finds cointegration vectors from the Johansen test used to form a mean-reverting portfolio.

        Note: Johansen test yields several linear combinations that may yield mean-reverting portfolios. The function
        stores all of them in decreasing order of eigenvalue meaning that the first linear combination forms the most
        mean-reverting portfolio which is used in trading. However, researchers may use other stored cointegration vectors
        to check other portfolios.

        This function will calculate and set johansen_trace_statistic and johansen_eigen_statistic only if
        the number of variables in the input dataframe is <=12. Otherwise it will generate a warning.

        A more detailed description of this method can be found on p. 54-58 of
        `"Algorithmic Trading: Winning Strategies and Their Rationale" by Ernie Chan
        <https://www.wiley.com/en-us/Algorithmic+Trading%3A+Winning+Strategies+and+Their+Rationale-p-9781118460146>`_.

        This function is a wrapper around the coint_johansen function from the statsmodels.tsa module. Detailed
        descriptions of this function are available in the
        `statsmodels documentation
        <https://www.statsmodels.org/stable/generated/statsmodels.tsa.vector_ar.vecm.coint_johansen.html>`_.

        :param price_data: (pd.DataFrame) Price data with columns containing asset prices.
        :param dependent_variable: (str) Column name which represents the dependent variable (y).
            By default, the first column is used as a dependent variable.
        :param det_order: (int) -1 for no deterministic term in Johansen test, 0 - for constant term, 1 - for linear trend.
        :param n_lags: (int) Number of lags used in the Johansen test. The practitioners use 1 as the default base value.
        """

        if not dependent_variable:
            dependent_variable = price_data.columns[0]

        self.price_data = price_data

        test_res = coint_johansen(price_data, det_order=det_order, k_ar_diff=n_lags)

        # Store eigenvectors in decreasing order of eigenvalues
        cointegration_vectors = pd.DataFrame(test_res.evec[:, test_res.ind].T, columns=price_data.columns)

        # Adjusting cointegration vectors to hedge ratios
        data_copy = price_data.copy()
        data_copy.drop(columns=dependent_variable, axis=1, inplace=True)

        # Store cointegration vectors
        self.cointegration_vectors = pd.DataFrame(test_res.evec[:, test_res.ind].T, columns=price_data.columns)

        # Calculating hedge ratios
        all_hedge_ratios = pd.DataFrame()

        # Convert to a format expected by `construct_spread` function and
        # normalize such that dependent has a hedge ratio 1.
        for vector in range(cointegration_vectors.shape[0]):

            hedge_ratios = cointegration_vectors.iloc[vector].to_dict()
            for ticker, ratio in hedge_ratios.items():
                if ticker != dependent_variable:
                    # Set value to be list to make it easier to read into pandas DataFrame
                    hedge_ratios[ticker] = [-ratio / hedge_ratios[dependent_variable]]
            # Set value to be list to make it easier to read into pandas DataFrame
            hedge_ratios[dependent_variable] = [1.0]

            # Concat together in one DataFrame
            all_hedge_ratios = pd.concat([all_hedge_ratios, pd.DataFrame(hedge_ratios)])

        self.hedge_ratios = all_hedge_ratios

        # Test critical values are available only if number of variables <= 12
        if price_data.shape[1] <= 12:
            # Eigenvalue test
            self.johansen_eigen_statistic = pd.DataFrame(test_res.max_eig_stat_crit_vals.T, columns=price_data.columns, index=["90%", "95%", "99%"])
            self.johansen_eigen_statistic.loc["eigen_value"] = test_res.max_eig_stat.T
            self.johansen_eigen_statistic.sort_index(ascending=False)

            # Trace statistic
            self.johansen_trace_statistic = pd.DataFrame(test_res.trace_stat_crit_vals.T, columns=price_data.columns, index=["90%", "95%", "99%"])
            self.johansen_trace_statistic.loc["trace_statistic"] = test_res.trace_stat.T
            self.johansen_trace_statistic.sort_index(ascending=False)


import numpy as np
from statsmodels.tsa.stattools import adfuller
from sklearn.linear_model import LinearRegression
from typing import Tuple


class EngleGrangerPortfolio(CointegratedPortfolio):
    """
    The class implements the construction of a mean-reverting portfolio using the two-step Engle-Granger method.
    It also tests model residuals for unit-root (presence of cointegration).
    """

    # pylint: disable=invalid-name
    def __init__(self):
        """
        Class constructor method.
        """

        self.price_data = None  # pd.DataFrame with price data used to fit the model.
        self.residuals = None  # OLS model residuals.
        self.dependent_variable = None  # Column name for dependent variable used in OLS estimation.
        self.cointegration_vectors = None  # Regression coefficients used as hedge-ratios.
        self.hedge_ratios = None  # Engle-Granger hedge ratios.
        self.adf_statistics = None  # ADF statistics.

    def perform_eg_test(self, residuals: pd.Series):
        """
        Perform Engle-Granger test on model residuals and generate test statistics and p values.

        :param residuals: (pd.Series) OLS residuals.
        """
        test_result = adfuller(residuals)
        critical_values = test_result[4]
        self.adf_statistics = pd.DataFrame(index=["99%", "95%", "90%"], data=critical_values.values())
        self.adf_statistics.loc["statistic_value", 0] = test_result[0]

    def fit(self, price_data: pd.DataFrame, add_constant: bool = False):
        """
        Finds hedge-ratios using a two-step Engle-Granger method to form a mean-reverting portfolio.
        By default, the first column of price data is used as a dependent variable in OLS estimation.

        This method was originally described in `"Co-integration and Error Correction: Representation,
        Estimation, and Testing," Econometrica, Econometric Society, vol. 55(2), pages 251-276, March 1987
        <https://www.jstor.org/stable/1913236>`_ by Engle, Robert F and Granger, Clive W J.

        :param price_data: (pd.DataFrame) Price data with columns containing asset prices.
        :param add_constant: (bool) A flag to add a constant term in linear regression.
        """

        self.price_data = price_data
        self.dependent_variable = price_data.columns[0]

        # Fit the regression
        hedge_ratios, _, _, residuals = self.get_ols_hedge_ratio(
            price_data=price_data, dependent_variable=self.dependent_variable, add_constant=add_constant
        )
        self.cointegration_vectors = pd.DataFrame(
            [np.append(1, -1 * np.array([hedge for ticker, hedge in hedge_ratios.items() if ticker != self.dependent_variable]))],
            columns=price_data.columns,
        )

        self.hedge_ratios = pd.DataFrame(
            [np.append(1, np.array([hedge for ticker, hedge in hedge_ratios.items() if ticker != self.dependent_variable]))],
            columns=price_data.columns,
        )

        # Get model residuals
        self.residuals = residuals
        self.perform_eg_test(self.residuals)

    @staticmethod
    def get_ols_hedge_ratio(
        price_data: pd.DataFrame, dependent_variable: str, add_constant: bool = False
    ) -> Tuple[dict, pd.DataFrame, pd.Series, pd.Series]:
        """
        Get OLS hedge ratio: y = beta*X.

        :param price_data: (pd.DataFrame) Data Frame with security prices.
        :param dependent_variable: (str) Column name which represents the dependent variable (y).
        :param add_constant: (bool) Boolean flag to add constant in regression setting.
        :return: (Tuple) Hedge ratios, X, and y and OLS fit residuals.
        """

        ols_model = LinearRegression(fit_intercept=add_constant)

        X = price_data.copy()
        X.drop(columns=dependent_variable, axis=1, inplace=True)
        exogenous_variables = X.columns.tolist()
        if X.shape[1] == 1:
            X = X.values.reshape(-1, 1)

        y = price_data[dependent_variable].copy()

        ols_model.fit(X, y)
        residuals = y - ols_model.predict(X)

        hedge_ratios = ols_model.coef_
        hedge_ratios_dict = dict(zip([dependent_variable] + exogenous_variables, np.insert(hedge_ratios, 0, 1.0)))

        return hedge_ratios_dict, X, y, residuals



import numpy as np


class DieboldLiCurve:
    def __init__(
        self, beta0: float, beta1: float, beta2: float, lambda_: float
    ) -> None:
        self.beta0 = beta0
        self.beta1 = beta1
        self.beta2 = beta2
        self.lambda_ = lambda_

    def yield_curve(self, tau: np.ndarray) -> np.ndarray:
        small_value = 1e-6  
        tau = np.maximum(tau, small_value)

        term1 = self.beta0
        term2 = self.beta1 * (1 - np.exp(-self.lambda_ * tau)) / (self.lambda_ * tau)
        term3 = self.beta2 * (
            (1 - np.exp(-self.lambda_ * tau)) / (self.lambda_ * tau)
            - np.exp(-self.lambda_ * tau)
        )
        return term1 + term2 + term3

    def __call__(self, tau: np.ndarray) -> np.ndarray:
        return self.yield_curve(tau)



import numpy as np
from scipy import optimize
from scipy.optimize import minimize
from scipy.optimize import check_grad
import pandas as pd
import random



"""
References:
https://ntnuopen.ntnu.no/ntnu-xmlui/bitstream/handle/11250/2738606/Wahlstr%C3%B8m2021_Article_AComparativeAnalysisOfParsimon.pdf?sequence=2
https://www.r-bloggers.com/2022/07/nelson-siegel-svensson-yield-curve-model-using-r-code/
https://deliverypdf.ssrn.com/delivery.php?ID=713094097085069095026119002078073069000039039014031001095005110028031090091006089071118026012125037127020071113073102127003097023054032039051005095126102077103088068084044104004103088089007127098090104125103094011087076120066067100084122003123097082&EXT=pdf&INDEX=TRUE
https://docs.scipy.org/doc/scipy/tutorial/optimize.html#sequential-least-squares-programming-slsqp-algorithm-method-slsqp
"""


class ZeroCouponCurve:
    
    """
    Estimate a zero coupon yield curve with the Nelson Siegel Svensson model
    
    """
    
    def __init__(self, yield_data, ytm_data, init_params, method = None,
                 fallback = None, percent = True, bounds = None, constraints = None):
        """
        Parameters
        ----------
        yield_data : DataFrame
            dataframe containing daily yield data
        ytm_data : DataFrame
            dataframe containing daily years to maturity data for each yield data point
            
        init_params : list
            list of initial parameter values
            
        percent : boolean
            if yield data are given in percent (ie. 3.51 for example instead of 0.0351) then True, otherwise False
            standard is True
            
        bounds : list
            bounds for parameter values in the optimization. list of pairwise values for each parameter
            if bounds = None, a standard set of parameters will be applied
        method : string
            main optimization problem to be used, default is SLSQP
        fallback : string
            alternative optimization method to be used if main method fails
        constraints : dict
            constraints for the optimization problem
        """
        
        self.yield_data = yield_data
        self.dates = self.yield_data['date']
        self.yield_data = self.yield_data.drop(columns = 'date')
        
        #Multiply ny 100 if yield data is not specified as percentage
        if percent == False:
            self.yield_data = np.array(self.yield_data) * 100
        else:
            self.yield_data = np.array(self.yield_data)
            
        
        self.ytm_data = ytm_data
        self.ytm_data = self.ytm_data.drop(columns = 'date')
        self.ytm_data = np.array(self.ytm_data)
        
        self.yield_data = self.yield_data.astype(float)
        self.ytm_data = self.ytm_data .astype(float)
        
        self.init_params = init_params
        
        if bounds == None:
            self.bounds = [(0,15),(-15,30),(-30,30),(-30,30),(0.001,30),(0.001,30)]
        else:
            self.bounds = bounds
            
        if constraints == None:
            self.cons=( 
                    {'type': 'ineq','fun': lambda x: x[0] + x[1]},
                    {'type': 'ineq','fun': lambda x: x[4]},
                    {'type': 'ineq','fun': lambda x: x[5]},
            )
        else:
            self.cons = constraints
        
        #If no optimization method is specified, use SLSQP as default
        if method == None:
            self.method = 'SLSQP'
        else:
            self.method = method
            
        #If no fallback optimization method is specified, use Nelder-Mead as default
        if fallback == None:
            self.fallback = 'Nelder-Mead'
        else:
            self.method = method

    def nss_model(self, T, params):
        """
        Return a yield for a given time to maturity (T) in years and a set of parameters
        """
        
        beta0 = params[0]
        beta1 = params[1]
        beta2 = params[2]
        beta3 = params[3]
        lambda0 = params[4]
        lambda1 = params[5]
        
        alpha1 = (1 - np.exp(-T / lambda0)) / (T / lambda0)
        alpha2 = (1 - np.exp(-T / lambda0)) / (T / lambda0) - np.exp(-T / lambda0)
        alpha3 = (1 - np.exp(-T / lambda1)) / (T / lambda1) - np.exp(-T / lambda1)
        
        return (beta0 + beta1 * alpha1 + beta2 * alpha2 + beta3 * alpha3)
    
    def nss_residuals(self, params, time_vec, yield_vec):
        """
        Parameters
        ----------
        params : list
            list of NSS parameters
        time_vec : list
            list of time to maturity values
        yield_vec : list
            list of yields corresponding to the given yield to maturity values
        """
        residuals = np.sum((self.nss_model(time_vec, params) - yield_vec)**2)
    
        return residuals
    

        
    def nss_optimize(self, params, time_vec, yield_vec, opt_method):
        """
        Parameters
        ----------
        params : numpy array
            array of initial parameter values
        time_vec : numpy array
            array of initial parameter values
        yield_vec : numpy array
            array of initial parameter values
        opt_method : string, optional
            set the optimization method. 
            The default is 'SLSQP'.
        """
        

        try:
            solution = minimize(self.nss_residuals, x0 = np.array(params), 
                                args = (time_vec, yield_vec), method = opt_method,
                                bounds = self.bounds,
                                constraints = self.cons)
            if solution.success:
                return solution.x, solution.fun
            else:
                return []
    
        except:
            pass
        
    def estimate_nss_params(self, mat_end, param_iter = 1):
        """
        estimate nss parameters for the whole dataset
        
        Parameters
        ----------
        mat_end : int
            last maturity to be estimated in months
        param_iter : int, optional
            Number of random trials for the optimization. If this is greater than one, the estimation process will
            select a random set of parameters and rerun the optimization. A the end, the parameter set
            with lowest residuals will be selected as the optimal set of parameters. The default is 1.

        """

        init_params = self.init_params
        results = np.arange(3,mat_end,1) / 12
        
        zc_rates = np.zeros((len(self.yield_data), len(results)))
        param_arr = np.zeros((len(self.yield_data), len(init_params)))
        resid_arr = []
        
        yields = self.yield_data
        ytm = self.ytm_data

        for i in range(0,len(yields)):
            print(i)
            y_vec = yields[i]
            t_vec = ytm[i]

            vec = np.stack((t_vec,y_vec)).T
            vec = vec[~np.isnan(vec[:,0])]
            vec = vec[~np.isnan(vec[:,1])]

            #Skip date if number of data points is lower than 4     
            if len(vec) < 4:
                pass
            else:
                vec = vec[vec[:,0].argsort()]
                
                #Adjust beta0 and beta1 to new set of yields
                beta0 = (vec[len(vec) - 1,1] + vec[len(vec) - 2,1]) / 2.0
                beta1 = vec[0][1] - beta0

                if len(init_params) > 0:
                    
                    init_params[0] = beta0
                    init_params[1] = beta1
                #If list of parameters from previous estimation is empty, choose the original initial parameters
                elif len(init_params) == 0:
                    init_params = self.init_params
                
                
                #Run through parameter iterations and get lists of residuals and tried parameters
                residuals, param_trials = self.run_param_iter(param_iter, init_params, vec)
                #Drop if residuals are zero
                residuals = residuals[~np.all(residuals == 0, axis = 1)]
                param_trials = param_trials[~np.all(param_trials == 0, axis = 1)]

                #Check if parameter list has any elements, skip if not
                if len(residuals) == 0:
                    pass
                else:
                    #Select the set of parameters with the lowest possible values
                    optim_params = param_trials[residuals.argmin()]
                    
                
                #Calculate a set of yields for the given parameters values
                res = self.nss_model(results, optim_params)
                param_arr[i] = optim_params
                zc_rates[i] = res
                resid_arr.append(residuals)
                
                #Set new parameter values to previously calculated optimal parameters
                init_params = optim_params
                
        return param_arr, zc_rates, resid_arr, vec
        
    def run_param_iter(self, param_iter, init_params, vec):
        
        residuals = np.zeros((param_iter,1))
        param_trials = np.zeros((param_iter, 6))
        param_trials[0] = init_params
        
        for j in range(param_iter):
            params = self.draw_new_init_params(init_params)
            #For the first estimation, use initial parameters init_params
            if j == 0:
                opt_res = self.nss_optimize(init_params, vec[:,0], vec[:,1], opt_method = self.method)
                #If minimize fails for the standard SLSQP method, try Nelder-Mead as a fallback
                if opt_res == None or len(opt_res) == 0:
                    print('Fallback-method used')
                    opt_res = self.nss_optimize(init_params, vec[:,0], vec[:,1], opt_method = self.fallback)
            #For subsequent parameter trials, use a randomly drawn parameter set
            else:
                opt_res = self.nss_optimize(params, vec[:,0], vec[:,1], opt_method = self.method)
                #If minimize fails for the standard SLSQP method, try Nelder-Mead as a fallback
                if opt_res == None or len(opt_res) == 0:
                    print('Fallback-method used')
                    opt_res = self.nss_optimize(init_params, vec[:,0], vec[:,1], opt_method = self.fallback)
            
            #If optimization still fails, then revert to the initial parameter set
            if opt_res == None or len(opt_res) == 0:
                print('Standard and fallback method failed, revert to initial parameter set')
                param_trials[j] = init_params
            
            else:
                param_trials[j] = opt_res[0]
                
                if opt_res[1] == 0:     #Set residuals to a very high value if they are zero. Zero residuals indicates a failed estimation
                    residuals[j] = 1000.0
                    print(residuals[j])
                else:
                            residuals[j] = opt_res[1]
                            
        return residuals, param_trials
                            
    def calc_nss_yields(self, parameters, results):
        
        yields = np.zeros((len(parameters), len(results)))
        
        for i in range(len(parameters)):
            yields[i] = self.nss_model(results, parameters[i])
            
        yields = pd.DataFrame(yields)
        yields = yields.loc[yields.sum(axis = 1) != 0]
        yields = pd.merge(self.dates, yields, left_on = self.dates.index, right_on = yields.index)
        yields = yields.drop(columns = 'key_0')
            
        return yields


    def draw_new_init_params(self, params):
        """
        Draw new initial parameters by randomly selecting another set of parameters
        beta0 and beta1 are selected as original values, but with a random adjustment
        """
        beta0 = params[0] + random.randint(-2,2)
        beta1 = params[1] + random.randint(-2,2)
        beta2 = random.randint(-30,30)
        beta3 = random.randint(-30,30)
        lambda0 = random.randint(0,30)
        lambda1 = random.randint(0,30)
        return [beta0, beta1, beta2, beta3, lambda0, lambda1]


import asyncio
import math
import os
import shutil
from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import datetime, timedelta
from itertools import product
from typing import Dict, List, Literal, Optional, Tuple, TypeAlias

import aiohttp
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import QuantLib as ql
import requests
import ujson as json
from scipy.optimize import minimize

JSON: TypeAlias = dict[str, "JSON"] | list["JSON"] | str | int | float | bool | None


def auction_df_filterer(historical_auctions_df: pd.DataFrame):
    historical_auctions_df = historical_auctions_df.copy()
    historical_auctions_df["issue_date"] = pd.to_datetime(
        historical_auctions_df["issue_date"]
        # , errors="coerce"
    )
    historical_auctions_df["maturity_date"] = pd.to_datetime(
        historical_auctions_df["maturity_date"]
        # , errors="coerce"
    )
    historical_auctions_df["auction_date"] = pd.to_datetime(
        historical_auctions_df["auction_date"]
        # , errors="coerce"
    )
    historical_auctions_df.loc[
        historical_auctions_df["original_security_term"].str.contains("29-Year", case=False, na=False),
        "original_security_term",
    ] = "30-Year"
    historical_auctions_df.loc[
        historical_auctions_df["original_security_term"].str.contains("30-", case=False, na=False),
        "original_security_term",
    ] = "30-Year"
    historical_auctions_df = historical_auctions_df[
        (historical_auctions_df["security_type"] == "Bill")
        | (historical_auctions_df["security_type"] == "Note")
        | (historical_auctions_df["security_type"] == "Bond")
    ]
    return historical_auctions_df


def build_treasurydirect_header(
    host_str: Optional[str] = "api.fiscaldata.treasury.gov",
    cookie_str: Optional[str] = None,
    origin_str: Optional[str] = None,
    referer_str: Optional[str] = None,
):
    return {
        "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7,application/json",
        "Accept-Encoding": "gzip, deflate, br, zstd",
        "Accept-Language": "en-US,en;q=0.9",
        "Cache-Control": "no-cache",
        "Connection": "keep-alive",
        "Cookie": cookie_str or "",
        "DNT": "1",
        "Host": host_str or "",
        "Origin": origin_str or "",
        "Referer": referer_str or "",
        "Pragma": "no-cache",
        "Sec-CH-UA": '"Not)A;Brand";v="99", "Google Chrome";v="127", "Chromium";v="127"',
        "Sec-CH-UA-Mobile": "?0",
        "Sec-CH-UA-Platform": '"Windows"',
        "Sec-Fetch-Dest": "document",
        "Sec-Fetch-Mode": "navigate",
        "Sec-Fetch-Site": "none",
        "Sec-Fetch-User": "?1",
        "Upgrade-Insecure-Requests": "1",
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/127.0.0.0 Safari/537.36",
    }


# n == 0 => On-the-runs
def get_last_n_off_the_run_cusips(
    auction_json: Optional[JSON] = None,
    auctions_df: Optional[pd.DataFrame] = None,
    n=0,
    filtered=False,
    as_of_date=datetime.today(),
    use_issue_date=False,
) -> List[Dict[str, str]] | pd.DataFrame:
    if not auction_json and auctions_df is None:
        return pd.DataFrame(columns=historical_auction_cols())

    if auction_json and auctions_df is None:
        auctions_df = pd.DataFrame(auction_json)

    auctions_df = auctions_df[
        (auctions_df["security_type"] != "TIPS")
        & (auctions_df["security_type"] != "TIPS Note")
        & (auctions_df["security_type"] != "TIPS Bond")
        & (auctions_df["security_type"] != "FRN")
        & (auctions_df["security_type"] != "FRN Note")
        & (auctions_df["security_type"] != "FRN Bond")
        & (auctions_df["security_type"] != "CMB")
    ]
    # auctions_df = auctions_df.drop(
    #     auctions_df[
    #         (auctions_df["security_type"] == "Bill")
    #         & (
    #             auctions_df["original_security_term"]
    #             != auctions_df["security_term_week_year"]
    #         )
    #     ].index
    # )
    auctions_df["auction_date"] = pd.to_datetime(auctions_df["auction_date"])
    auctions_df["issue_date"] = pd.to_datetime(auctions_df["issue_date"])
    current_date = as_of_date
    auctions_df = auctions_df[auctions_df["auction_date" if not use_issue_date else "issue_date"].dt.date <= current_date.date()]
    auctions_df = auctions_df.sort_values("auction_date" if not use_issue_date else "issue_date", ascending=False)

    mapping = {
        "4-Week": 0.077,
        "8-Week": 0.15,
        "13-Week": 0.25,
        "17-Week": 0.33,
        "26-Week": 0.5,
        "52-Week": 1,
        "2-Year": 2,
        "3-Year": 3,
        "5-Year": 5,
        "7-Year": 7,
        "10-Year": 10,
        "20-Year": 20,
        "30-Year": 30,
    }

    on_the_run = auctions_df.groupby("original_security_term").first().reset_index()
    on_the_run = on_the_run[(on_the_run["security_type"] == "Note") | (on_the_run["security_type"] == "Bond")]
    on_the_run_result = on_the_run[
        [
            "original_security_term",
            "security_type",
            "cusip",
            "auction_date",
            "issue_date",
        ]
    ]

    on_the_run_bills = auctions_df.groupby("security_term").first().reset_index()
    on_the_run_bills = on_the_run_bills[on_the_run_bills["security_type"] == "Bill"]
    on_the_run_result_bills = on_the_run_bills[
        [
            "original_security_term",
            "security_type",
            "cusip",
            "auction_date",
            "issue_date",
        ]
    ]

    on_the_run = pd.concat([on_the_run_result_bills, on_the_run_result])

    if n == 0:
        return on_the_run

    off_the_run = auctions_df[~auctions_df.index.isin(on_the_run.index)]
    off_the_run_result = off_the_run.groupby("original_security_term").nth(list(range(1, n + 1))).reset_index()

    combined_result = pd.concat([on_the_run_result, off_the_run_result], ignore_index=True)
    combined_result = combined_result.sort_values(by=["original_security_term", "issue_date"], ascending=[True, False])

    combined_result["target_tenor"] = combined_result["original_security_term"].replace(mapping)
    mask = combined_result["original_security_term"].isin(mapping.keys())
    mapped_and_filtered_df = combined_result[mask]
    grouped = mapped_and_filtered_df.groupby("original_security_term")
    max_size = grouped.size().max()
    wrapper = []
    for i in range(max_size):
        sublist = []
        for _, group in grouped:
            if i < len(group):
                sublist.append(group.iloc[i].to_dict())
        sublist = sorted(sublist, key=lambda d: d["target_tenor"])
        if filtered:
            wrapper.append({auctioned_dict["target_tenor"]: auctioned_dict["cusip"] for auctioned_dict in sublist})
        else:
            wrapper.append(sublist)

    return wrapper


def get_historical_on_the_run_cusips(
    auctions_df: pd.DataFrame,
    as_of_date=datetime.today(),
    use_issue_date=False,
) -> pd.DataFrame:

    current_date = as_of_date
    auctions_df = auctions_df[auctions_df["auction_date" if not use_issue_date else "issue_date"].dt.date <= current_date.date()]
    auctions_df = auctions_df[auctions_df["maturity_date"].dt.date >= current_date.date()]
    auctions_df = auctions_df.sort_values("auction_date" if not use_issue_date else "issue_date", ascending=False)

    mapping = {
        "17-Week": 0.25,
        "26-Week": 0.5,
        "52-Week": 1,
        "2-Year": 2,
        "3-Year": 3,
        "5-Year": 5,
        "7-Year": 7,
        "10-Year": 10,
        "20-Year": 20,
        "30-Year": 30,
    }

    on_the_run_df = auctions_df.groupby("original_security_term").first().reset_index()
    on_the_run_filtered_df = on_the_run_df[
        [
            "original_security_term",
            "security_type",
            "cusip",
            "auction_date",
            "issue_date",
        ]
    ]
    on_the_run_filtered_df["target_tenor"] = on_the_run_filtered_df["original_security_term"].replace(mapping)

    return on_the_run_filtered_df


def get_active_cusips(
    auction_json: Optional[JSON] = None,
    historical_auctions_df: Optional[pd.DataFrame] = None,
    as_of_date=datetime.today(),
    use_issue_date=False,
) -> pd.DataFrame:
    if not auction_json and historical_auctions_df is None:
        return pd.DataFrame(columns=historical_auction_cols())

    if auction_json and historical_auctions_df is None:
        historical_auctions_df = pd.DataFrame(auction_json)

    historical_auctions_df = auction_df_filterer(historical_auctions_df)
    historical_auctions_df = historical_auctions_df[
        historical_auctions_df["auction_date" if not use_issue_date else "issue_date"].dt.date <= as_of_date.date()
    ]
    historical_auctions_df = historical_auctions_df[historical_auctions_df["maturity_date"] >= as_of_date]
    historical_auctions_df = historical_auctions_df.drop_duplicates(subset=["cusip"], keep="first")
    historical_auctions_df["int_rate"] = pd.to_numeric(historical_auctions_df["int_rate"], errors="coerce")
    historical_auctions_df["time_to_maturity"] = (historical_auctions_df["maturity_date"] - as_of_date).dt.days / 365
    return historical_auctions_df


def last_day_n_months_ago(given_date: datetime, n: int = 1, return_all: bool = False) -> datetime | List[datetime]:
    if return_all:
        given_date = pd.Timestamp(given_date)
        return [(given_date - pd.offsets.MonthEnd(i)).to_pydatetime() for i in range(1, n + 1)]

    given_date = pd.Timestamp(given_date)
    last_day = given_date - pd.offsets.MonthEnd(n)
    return last_day.to_pydatetime()


def cookie_string_to_dict(cookie_string):
    cookie_pairs = cookie_string.split("; ")
    cookie_dict = {pair.split("=")[0]: pair.split("=")[1] for pair in cookie_pairs if "=" in pair}
    return cookie_dict


def is_valid_ust_cusip(potential_ust_cusip: str):
    return len(potential_ust_cusip) == 9 and "912" in potential_ust_cusip


def historical_auction_cols():
    return [
        "cusip",
        "security_type",
        "auction_date",
        "issue_date",
        "maturity_date",
        "price_per100",
        "allocation_pctage",
        "avg_med_yield",
        "bid_to_cover_ratio",
        "comp_accepted",
        "comp_tendered",
        "corpus_cusip",
        "tint_cusip_1",
        "currently_outstanding",
        "direct_bidder_accepted",
        "direct_bidder_tendered",
        "est_pub_held_mat_by_type_amt",
        "fima_included",
        "fima_noncomp_accepted",
        "fima_noncomp_tendered",
        "high_discnt_rate",
        "high_investment_rate",
        "high_price",
        "high_yield",
        "indirect_bidder_accepted",
        "indirect_bidder_tendered",
        "int_rate",
        "low_investment_rate",
        "low_price",
        "low_discnt_margin",
        "low_yield",
        "max_comp_award",
        "max_noncomp_award",
        "noncomp_accepted",
        "noncomp_tenders_accepted",
        "offering_amt",
        "security_term",
        "original_security_term",
        "security_term_week_year",
        "primary_dealer_accepted",
        "primary_dealer_tendered",
        "reopening",
        "total_accepted",
        "total_tendered",
        "treas_retail_accepted",
        "treas_retail_tenders_accepted",
    ]


def ust_labeler(row: pd.Series):
    mat_date = row["maturity_date"]
    tenor = row["original_security_term"]
    if np.isnan(row["int_rate"]):
        return str(row["high_investment_rate"])[:5] + "s , " + mat_date.strftime("%b %y") + "s" + ", " + tenor
    return str(row["int_rate"]) + "s, " + mat_date.strftime("%b %y") + "s, " + tenor


def ust_sorter(term: str):
    if " " in term:
        term = term.split(" ")[0]
    num, unit = term.split("-")
    num = int(num)
    unit_multiplier = {"Year": 365, "Month": 30, "Week": 7, "Day": 1}
    return num * unit_multiplier[unit]


def get_otr_cusips_by_date(historical_auctions_df: pd.DataFrame, dates: list, use_issue_date: bool = True):
    historical_auctions_df = auction_df_filterer(historical_auctions_df)
    date_column = "issue_date" if use_issue_date else "auction_date"
    historical_auctions_df = historical_auctions_df.sort_values(by=[date_column], ascending=False)
    historical_auctions_df = historical_auctions_df.drop_duplicates(subset=["cusip"], keep="last")
    grouped = historical_auctions_df.groupby("original_security_term")
    otr_cusips_by_date = {date: [] for date in dates}
    for _, group in grouped:
        group = group.reset_index(drop=True)
        for date in dates:
            filtered_group = group[(group[date_column] <= date) & (group["maturity_date"] > date)]
            if not filtered_group.empty:
                otr_cusip = filtered_group.iloc[0]["cusip"]
                otr_cusips_by_date[date].append(otr_cusip)

    return otr_cusips_by_date


def process_cusip_otr_daterange(cusip, historical_auctions_df, date_column):
    try:
        tenor = historical_auctions_df[historical_auctions_df["cusip"] == cusip]["original_security_term"].iloc[0]
        tenor_df: pd.DataFrame = historical_auctions_df[historical_auctions_df["original_security_term"] == tenor].reset_index()
        otr_df = tenor_df[tenor_df["cusip"] == cusip]
        otr_index = otr_df.index[0]
        start_date: pd.Timestamp = otr_df[date_column].iloc[0]
        start_date = start_date.to_pydatetime()

        if otr_index == 0:
            return cusip, (start_date, datetime.today().date())

        if otr_index < len(tenor_df) - 1:
            end_date: pd.Timestamp = tenor_df[date_column].iloc[otr_index - 1]
            end_date = end_date.to_pydatetime()
        else:
            end_date = datetime.today().date()

        return cusip, {"start_date": start_date, "end_date": end_date}
    except Exception as e:
        # print(f"Something went wrong for {cusip}: {e}")
        return cusip, {"start_date": None, "end_date": None}


def get_otr_date_ranges(historical_auctions_df: pd.DataFrame, cusips: List[str], use_issue_date: bool = True) -> Dict[str, Tuple[datetime, datetime]]:

    historical_auctions_df = auction_df_filterer(historical_auctions_df)
    date_column = "issue_date" if use_issue_date else "auction_date"
    historical_auctions_df = historical_auctions_df.sort_values(by=[date_column], ascending=False)
    historical_auctions_df = historical_auctions_df[historical_auctions_df["issue_date"].dt.date < datetime.today().date()]
    historical_auctions_df = historical_auctions_df.drop_duplicates(subset=["cusip"], keep="last")

    cusip_daterange_map = {}
    with ThreadPoolExecutor() as executor:
        futures = {executor.submit(process_cusip_otr_daterange, cusip, historical_auctions_df, date_column): cusip for cusip in cusips}

        for future in as_completed(futures):
            cusip, date_range = future.result()
            cusip_daterange_map[cusip] = date_range

    return cusip_daterange_map


def pydatetime_to_quantlib_date(py_datetime: datetime) -> ql.Date:
    return ql.Date(py_datetime.day, py_datetime.month, py_datetime.year)


def quantlib_date_to_pydatetime(ql_date: ql.Date):
    return datetime(ql_date.year(), ql_date.month(), ql_date.dayOfMonth())


def get_isin_from_cusip(cusip_str, country_code: str = "US"):
    """
    >>> get_isin_from_cusip('037833100', 'US')
    'US0378331005'
    """
    isin_to_digest = country_code + cusip_str.upper()

    get_numerical_code = lambda c: str(ord(c) - 55)
    encode_letters = lambda c: c if c.isdigit() else get_numerical_code(c)
    to_digest = "".join(map(encode_letters, isin_to_digest))

    ints = [int(s) for s in to_digest[::-1]]
    every_second_doubled = [x * 2 for x in ints[::2]] + ints[1::2]

    sum_digits = lambda i: sum(divmod(i, 10))
    digit_sum = sum([sum_digits(i) for i in every_second_doubled])

    check_digit = (10 - digit_sum % 10) % 10
    return isin_to_digest + str(check_digit)


def get_cstrips_cusips(
    historical_auctions_df: pd.DataFrame,
    as_of_date: Optional[datetime] = None,
):
    historical_auctions_df = auction_df_filterer(historical_auctions_df)
    active_df = historical_auctions_df[historical_auctions_df["maturity_date"] > as_of_date]
    tint_cusip = "tint_cusip_1"
    active_df[tint_cusip] = active_df[tint_cusip].replace("null", np.nan)
    active_df = active_df[active_df[tint_cusip].notna()]
    active_df = active_df.sort_values(by=["maturity_date"]).reset_index(drop=True)
    return active_df[["maturity_date", tint_cusip]]


def original_security_term_to_ql_period(original_security_term: str):
    ost_ql_period = {
        "52-Week": ql.Period("1Y"),
        "2-Year": ql.Period("2Y"),
        "3-Year": ql.Period("3Y"),
        "5-Year": ql.Period("5Y"),
        "7-Year": ql.Period("7Y"),
        "10-Year": ql.Period("10Y"),
        "20-Year": ql.Period("20Y"),
        "30-Year": ql.Period("30Y"),
    }
    return ost_ql_period[original_security_term]


def enhanced_plotly_blue_scale():
    enhanced_blue_scale = [
        [0.0, "#0508b8"],  # Deepest blue
        [0.02, "#0508b8"],  # Extending deepest blue
        [0.04, "#0610b9"],  # Very subtle shift
        [0.06, "#0712ba"],  # Slightly lighter
        [0.08, "#0814bb"],  # Incrementally lighter
        [0.10, "#0916bc"],  # Another slight lightening
        [0.12, "#0a18bd"],  # Continuing the trend
        [0.14, "#0b1abd"],  # Gradual lightening
        [0.16, "#0c1cbe"],  # More noticeable change
        [0.18, "#0d1ebe"],  # Further lightening
        [0.20, "#0e20bf"],  # Approaching mid blues
        [0.25, "#1116e8"],  # Jumping to a lighter blue for contrast
        [0.3, "#1910d8"],  # Original Plotly3 blue
    ]

    # Transitioning back to the original Plotly3 scale, starting from a point after the enhanced blue
    transition_scale = [
        [0.35, "#3c19f0"],
        [0.40, "#6b1cfb"],
        [0.45, "#981cfd"],
        [0.50, "#bf1cfd"],
        [0.55, "#dd2bfd"],
        [0.60, "#f246fe"],
        [0.65, "#fc67fd"],
        [0.70, "#fe88fc"],
        [0.75, "#fea5fd"],
        [0.80, "#febefe"],
        [0.85, "#fec3fe"],
    ]

    # Combine the more granular blue scale with the rest of the Plotly3 colors
    combined_scale = enhanced_blue_scale + transition_scale

    return combined_scale


def to_quantlib_fixed_rate_bond_obj(bond_info: Dict[str, str], as_of_date: datetime, print_bond_info=False):
    if print_bond_info:
        print(json.dumps(bond_info, indent=4, default=str))
    maturity_date: pd.Timestamp = bond_info["maturity_date"]
    calendar = ql.UnitedStates(m=ql.UnitedStates.GovernmentBond)
    today = calendar.adjust(pydatetime_to_quantlib_date(py_datetime=as_of_date))
    ql.Settings.instance().evaluationDate = today
    t_plus = 1
    bond_settlement_date = calendar.advance(today, ql.Period(t_plus, ql.Days))

    schedule = ql.Schedule(
        bond_settlement_date,
        pydatetime_to_quantlib_date(maturity_date),
        ql.Period(ql.Semiannual),
        calendar,
        ql.ModifiedFollowing,
        ql.ModifiedFollowing,
        ql.DateGeneration.Backward,
        False,
    )
    return ql.FixedRateBond(
        t_plus,
        100.0,
        schedule,
        [bond_info["int_rate"] / 100],
        ql.ActualActual(ql.ActualActual.ISDA),
    )


class NoneReturningSpline:
    def __init__(self, *args, **kwargs):
        # Accept the same parameters as a usual spline, but do nothing with them
        pass

    def __call__(self, x):
        # Always return None for any input x
        if isinstance(x, (np.ndarray, list)):
            return [None] * len(x)  # Return a list of None for multiple inputs
        else:
            return None



# adapted from: https://github.com/ahgperrin/PyCurve/blob/master/src/PyCurve/simulation.py

from typing import Any, Tuple

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt


class Simulation:
    def __init__(self, simulated_paths: np.ndarray, dt: float) -> None:
        self._sim = self._is_valid_attr(simulated_paths)
        self._dt = dt

    @property
    def get_sim(self) -> np.ndarray:
        return self.__getattribute__("_sim")

    @property
    def get_nb_sim(self) -> int:
        return self.__getattribute__("_sim").shape[1]

    @property
    def get_steps(self) -> int:
        return self.__getattribute__("_sim").shape[0]

    @property
    def get_dt(self) -> float:
        return self.__getattribute__("_dt")

    @staticmethod
    def _is_valid_attr(attr: Any) -> np.ndarray:
        assert isinstance(
            attr, np.ndarray
        ), "Class Constructor takes only numpy arrays or list as arguments"
        return attr

    def yield_curve(self) -> Tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
        discount_factor: np.ndarray = self.discount_factor()
        yield_curve = (
            np.mean(discount_factor, axis=1)
            ** (-1 / np.full(self.get_steps, self.get_dt).cumsum())
        ) - 1
        return np.full(self.get_steps, self.get_dt).cumsum(), yield_curve

    def discount_factor(self) -> np.array:
        rieman_sum: np.ndarray = np.zeros(shape=self.get_sim.shape)
        discount_factor: np.ndarray = np.zeros(shape=self.get_sim.shape)
        for sim in range(self.get_nb_sim):
            rieman_sum[:, sim] = np.cumsum(
                np.multiply(self.get_sim[:, sim], self.get_dt)
            )
            discount_factor[:, sim] = np.exp(-rieman_sum[:, sim])
        return discount_factor

    def plot_discount_curve(self, average: bool = False) -> None:
        discount_factor: np.ndarray = self.discount_factor()
        t: np.ndarray = np.full(self.get_steps, self.get_dt).cumsum()
        fig, ax = plt.subplots(1)
        fig.canvas.set_window_title("Discount Factor")
        fig.suptitle("Discount Factor")
        ax.set_xlabel("Time, t")
        ax.set_ylabel("Simulated Discount Factor")
        if average:
            ax.plot(t, np.mean(discount_factor, axis=1), c="navy")
        else:
            ax.plot(t, discount_factor)

    def plot_simulation(self) -> None:
        t: np.ndarray = np.full(self.get_steps, self.get_dt).cumsum()
        fig, ax = plt.subplots(1)
        fig.suptitle("Simulated Paths")
        fig.canvas.set_window_title("Simulated Paths")
        ax.set_xlabel("Time, t")
        ax.set_ylabel("Simulated Yield")
        ax.plot(t, self.get_sim, lw=0.5)
        plt.show()
        return fig



from datetime import datetime
from typing import Dict, List, Callable, Optional, Tuple

import matplotlib.pylab as pylab
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy
import scipy.interpolate
import tqdm

params = {
    "axes.titlesize": "x-large",
    "legend.fontsize": "x-large",
    "axes.labelsize": "x-large",
    "xtick.labelsize": "x-large",
    "ytick.labelsize": "x-large",
}
pylab.rcParams.update(params)

import seaborn as sns

from CurvyCUSIPs.CurveDataFetcher import CurveDataFetcher
from CurvyCUSIPs.CurveInterpolator import GeneralCurveInterpolator
from CurvyCUSIPs.CurveBuilder import calc_ust_metrics
from CurvyCUSIPs.utils.ust_viz import plot_usts
from CurvyCUSIPs.utils.regression_utils import run_basic_linear_regression, run_basic_linear_regression_df, plot_residuals_timeseries, run_rolling_regression_df

sns.set(style="whitegrid", palette="dark")

import warnings

warnings.filterwarnings("ignore", category=FutureWarning)
warnings.filterwarnings("ignore", category=pd.errors.SettingWithCopyWarning)


def cusip_spread_rv_regression(
    curve_data_fetcher: CurveDataFetcher,
    label1: str,
    label2: str,
    cusip_timeseries: Dict[str, List[Dict[str, str | float | int]]],
    fitted_splines_timeseries_dict: Dict[str, Dict[datetime, Callable]],
    on_rate: float,
    benchmark_tenor_1: Optional[int] = None,
    benchmark_tenor_2: Optional[int] = None,
    date_subset: Optional[Tuple[datetime, datetime]] = None,
    main_spline_key: Optional[str] = None,
    ct_yields_df: Optional[pd.DataFrame] = None,
    # rolling_r2_window: Optional[int] = 60,
):
    if not main_spline_key:
        main_spline_key = list(fitted_splines_timeseries_dict.keys())[0]

    cusip1 = curve_data_fetcher.ust_data_fetcher.ust_label_to_cusip(label1)["cusip"]
    cusip2 = curve_data_fetcher.ust_data_fetcher.ust_label_to_cusip(label2)["cusip"]
    cusip1_df = pd.DataFrame(cusip_timeseries[cusip1]).sort_values(by=["Date"])
    cusip2_df = pd.DataFrame(cusip_timeseries[cusip2]).sort_values(by=["Date"])
    if date_subset:
        cusip1_df = cusip1_df[(cusip1_df["Date"] >= date_subset[0]) & (cusip1_df["Date"] <= date_subset[1])]
        cusip2_df = cusip2_df[(cusip2_df["Date"] >= date_subset[0]) & (cusip2_df["Date"] <= date_subset[1])]

    if benchmark_tenor_1 and benchmark_tenor_2:
        benchmark_spline_spreads: Dict[str, List[float]] = {}
        for spline_key, splines_timeseries in fitted_splines_timeseries_dict.items():
            for dt in cusip1_df["Date"].to_list():
                dt: pd.Timestamp = dt
                curr_spread = splines_timeseries[dt.to_pydatetime()](benchmark_tenor_2) - splines_timeseries[dt.to_pydatetime()](benchmark_tenor_1)
                if spline_key not in benchmark_spline_spreads:
                    benchmark_spline_spreads[spline_key] = []

                benchmark_spline_spreads[spline_key].append(curr_spread)

    spread_dict = {
        "Date": cusip1_df["Date"],
        label1: cusip1_df["eod_yield"],
        label2: cusip2_df["eod_yield"],
        f"{label1} / {label2}": cusip2_df["eod_yield"] - cusip1_df["eod_yield"],
        f"{label1}_free_float": cusip1_df["free_float"],
        f"{label2}_free_float": cusip2_df["free_float"],
        f"{label1}_est_outstanding_amount": cusip1_df["est_outstanding_amount"],
        f"{label1}_soma_holdings": cusip1_df["soma_holdings"],
        f"{label1}_soma_holdings_percent_outstanding": cusip1_df["soma_holdings_percent_outstanding"],
        f"{label1}_stripped_amount": cusip1_df["stripped_amount"],
        f"{label1}_reconstituted_amount": cusip1_df["reconstituted_amount"],
        f"{label2}_est_outstanding_amount": cusip2_df["est_outstanding_amount"],
        f"{label2}_soma_holdings": cusip2_df["soma_holdings"],
        f"{label2}_soma_holdings_percent_outstanding": cusip2_df["soma_holdings_percent_outstanding"],
        f"{label2}_stripped_amount": cusip2_df["stripped_amount"],
        f"{label2}_reconstituted_amount": cusip2_df["reconstituted_amount"],
    }
    for spline_key in fitted_splines_timeseries_dict.keys():
        spread_dict[f"{label1}_{spline_key}_spread"] = cusip1_df[f"{spline_key}_spread"]
        spread_dict[f"{label2}_{spline_key}_spread"] = cusip2_df[f"{spline_key}_spread"]
        if benchmark_tenor_1 and benchmark_tenor_2:
            spread_dict[f"{spline_key}_{benchmark_tenor_1}s{benchmark_tenor_2}s"] = benchmark_spline_spreads[spline_key]

    spread_df = pd.DataFrame(spread_dict)
    # f"{main_spline_key}_{benchmark_tenor_1}s{benchmark_tenor_2}s"

    plt.figure(figsize=(20, 10))
    plt.plot(spread_df["Date"], spread_df[label1], label=label1)
    plt.plot(spread_df["Date"], spread_df[label2], label=label2)
    plt.xlabel("Date")
    plt.ylabel("Yield")
    plt.title(f"{label1} vs {label2}", fontdict={"fontsize": "x-large"})
    plt.legend(fontsize="x-large")
    plt.grid(True)
    plt.show()

    for curr_label in [label1, label2]:
        plt.figure(figsize=(20, 10))
        for spline_key in fitted_splines_timeseries_dict.keys():
            curr_spline_spread = spread_df[f"{curr_label}_{spline_key}_spread"]
            plt.plot(spread_df["Date"], curr_spline_spread, label=f"{curr_label} {spline_key} Spread")

            if spline_key == main_spline_key:
                spline_spread_mean = curr_spline_spread.mean()
                spline_spread_std = curr_spline_spread.std()
                plt.axhline(spline_spread_mean, color="green", linestyle="--", label=f"Mean ({curr_label} {main_spline_key} Spread)")
                plt.axhline(0, color="red", linestyle="--")

                for std in [1, 2]:
                    upper = spline_spread_mean + spline_spread_std * std
                    lower = spline_spread_mean - spline_spread_std * std
                    curr = plt.axhline(upper, linestyle="--", label=f"+/- {std} STD ({curr_label} {main_spline_key} Spread)")
                    plt.axhline(lower, linestyle="--", color=curr.get_color())

        plt.xlabel("Date")
        plt.ylabel("Spread (bps)")
        plt.title(f"{curr_label} Spline Spreads", fontdict={"fontsize": "x-large"})
        plt.legend(fontsize="x-large")
        plt.grid(True)
        plt.show()

    plt.figure(figsize=(20, 10))
    plt.plot(spread_df["Date"], spread_df[f"{label1} / {label2}"], label=f"{label1} / {label2}")
    plt.xlabel("Date")
    plt.ylabel("Yield (bps)")
    plt.title(f"{label1} / {label2} Spread", fontdict={"fontsize": "x-large"})
    plt.legend(fontsize="x-large")
    plt.grid(True)
    plt.show()

    if benchmark_tenor_1 and benchmark_tenor_2:
        print(f"{main_spline_key} is Benchmark Spline")
        plt.figure(figsize=(20, 10))
        plt.plot(
            spread_df["Date"],
            spread_df[f"{main_spline_key}_{benchmark_tenor_1}s{benchmark_tenor_2}s"],
            label=f"{benchmark_tenor_1}s{benchmark_tenor_2}s",
        )
        plt.xlabel("Date")
        plt.ylabel("Yield (bps)")
        plt.title(f"{benchmark_tenor_1}s{benchmark_tenor_2}s Benchmark curve", fontdict={"fontsize": "x-large"})
        plt.legend(fontsize="x-large")
        plt.grid(True)
        plt.show()

    for liquidity_metric in [
        "free_float",
        "est_outstanding_amount",
        "soma_holdings",
        "soma_holdings_percent_outstanding",
        "stripped_amount",
        "reconstituted_amount",
    ]:
        fig, ax1 = plt.subplots(figsize=(20, 10))
        (line1,) = ax1.plot(spread_df["Date"], spread_df[f"{label1}_{liquidity_metric}"], label=f"{label1}_{liquidity_metric}", color="b")
        ax1.set_xlabel("Date")
        ax1.set_ylabel(f"{label1} {liquidity_metric} ($mm)", color="b", fontsize=18)
        ax1.tick_params(axis="y", labelcolor="b", labelsize=14)
        ax1.grid(True)
        ax2 = ax1.twinx()
        (line2,) = ax2.plot(spread_df["Date"], spread_df[f"{label2}_{liquidity_metric}"], label=f"{label2}_{liquidity_metric}", color="r")
        ax2.set_ylabel(f"{label2} {liquidity_metric} ($mm)", color="r", fontsize=18)
        ax2.tick_params(axis="y", labelcolor="r", labelsize=14)
        plt.title(f"{label1} vs {label2} {liquidity_metric}", fontsize="x-large")
        plt.legend([line1, line2], [line1.get_label(), line2.get_label()], fontsize="x-large")
        fig.tight_layout()
        plt.show()

    if benchmark_tenor_1 and benchmark_tenor_2:
        r = run_basic_linear_regression_df(
            df=spread_df,
            x_col=f"{main_spline_key}_{benchmark_tenor_1}s{benchmark_tenor_2}s",
            y_col=f"{label1} / {label2}",
        )
        plot_residuals_timeseries(df=spread_df, results=r, stds=[1, 2])
        # plot_residuals_timeseries(df=spread_df, results=r, plot_zscores=True)

    if ct_yields_df is not None:
        ct_yields_df[f"CT{benchmark_tenor_1}-CT{benchmark_tenor_2}"] = ct_yields_df[f"CT{benchmark_tenor_2}"] - ct_yields_df[f"CT{benchmark_tenor_1}"]
        ct_spread_df = pd.merge(left=spread_df[["Date", f"{label1} / {label2}"]], right=ct_yields_df, on="Date", how="inner")
        r = run_basic_linear_regression_df(
            df=ct_spread_df,
            x_col=f"CT{benchmark_tenor_1}-CT{benchmark_tenor_2}",
            y_col=f"{label1} / {label2}",
        )
        plot_residuals_timeseries(df=ct_spread_df, results=r, stds=[1, 2])

    # run_rolling_regression_df(
    #     df=spread_df, x_col=f"{main_spline_key}_{benchmark_tenor_1}s{benchmark_tenor_2}s", y_col=f"{label1} / {label2}", window=rolling_r2_window
    # )

    interp_func_key = list(fitted_splines_timeseries_dict.keys())[0]
    print(f"Using {interp_func_key} for UST Metrics Calcs")
    cusip1_metrics = []
    for _, row in tqdm.tqdm(cusip1_df.iterrows(), desc=f"{label1} Metrics Calc"):
        cusip1_metrics.append(
            calc_ust_metrics(
                bond_info=curve_data_fetcher.ust_data_fetcher.cusip_to_ust_label(cusip=row["cusip"]),
                curr_price=row["eod_price"],
                curr_ytm=row["eod_yield"],
                as_of_date=row["Date"],
                scipy_interp=row[interp_func_key],
                on_rate=on_rate,
            )
        )
    cusip1_metrics_df = pd.DataFrame(cusip1_metrics)

    cusip2_metrics = []
    for _, row in tqdm.tqdm(cusip2_df.iterrows(), desc=f"{label2} Metrics Calc"):
        cusip2_metrics.append(
            calc_ust_metrics(
                bond_info=curve_data_fetcher.ust_data_fetcher.cusip_to_ust_label(cusip=row["cusip"]),
                curr_price=row["eod_price"],
                curr_ytm=row["eod_yield"],
                as_of_date=row["Date"],
                scipy_interp=row[interp_func_key],
                on_rate=on_rate,
            )
        )
    cusip2_metrics_df = pd.DataFrame(cusip2_metrics)

    plt.figure(figsize=(20, 10))
    plt.plot(cusip1_metrics_df["Date"], cusip1_metrics_df["zspread"], label=f"{label1} Z-Spread")
    plt.plot(cusip2_metrics_df["Date"], cusip2_metrics_df["zspread"], label=f"{label2} Z-Spread")
    plt.xlabel("Date")
    plt.ylabel("Yield (bps)")
    plt.title(f"{label1} vs {label2} Z-Spreads", fontdict={"fontsize": "x-large"})
    plt.legend(fontsize="x-large")
    plt.grid(True)
    plt.show()

    # TODO FIXED THIS
    # plt.figure(figsize=(20, 10))
    # plt.plot(cusip1_metrics_df["Date"], cusip1_metrics_df["3m_carry_and_roll"], label=f"{label1} 3m Carry & Roll")
    # plt.plot(cusip1_metrics_df["Date"], cusip1_metrics_df["12m_carry_and_roll"], label=f"{label1} 12m Carry & Roll")
    # plt.plot(cusip2_metrics_df["Date"], cusip2_metrics_df["3m_carry_and_roll"], label=f"{label2} 3m Carry & Roll")
    # plt.plot(cusip2_metrics_df["Date"], cusip2_metrics_df["12m_carry_and_roll"], label=f"{label2} 12m Carry & Roll")
    # plt.xlabel("Date")
    # plt.ylabel("Yield (bps)")
    # plt.title(f"{label1} vs {label2} Carry & Roll Profiles", fontdict={"fontsize": "x-large"})
    # plt.legend(fontsize="x-large")
    # plt.grid(True)
    # plt.show()

    plt.figure(figsize=(20, 10))
    clean_price_spread_1 = cusip1_metrics_df["clean_price"] - cusip1_metrics_df["zspread_impl_clean_price"]
    clean_price_spread_2 = cusip2_metrics_df["clean_price"] - cusip2_metrics_df["zspread_impl_clean_price"]
    plt.plot(cusip1_metrics_df["Date"], clean_price_spread_1, label=f"{label1} Clean Price Spread")
    plt.plot(cusip1_metrics_df["Date"], clean_price_spread_2, label=f"{label2} Clean Price Spread")
    plt.xlabel("Date")
    plt.ylabel("Yield (bps)")
    plt.title(f"{label1} vs {label2} Clean Price vs Z-Spread Implied Clean Price Spread", fontdict={"fontsize": "x-large"})
    plt.legend(fontsize="x-large")
    plt.grid(True)
    plt.show()



# Stolen from https://github.com/antdvid/MonotonicCubicInterpolation/blob/master/monospline.py


import numpy as np


class MonoSpline:
    def __init__(self, x, y):
        self.x = np.array(x)
        self.y = np.array(y)
        self.n = self.y.size
        self.h = self.x[1:] - self.x[:-1]
        self.m = (self.y[1:] - self.y[:-1]) / self.h
        self.a = self.y[:]
        self.b = self.compute_b(self.x, self.y)
        self.c = (3 * self.m - self.b[1:] - 2 * self.b[:-1]) / self.h
        self.d = (self.b[1:] + self.b[:-1] - 2 * self.m) / (self.h * self.h)

    def compute_b(self, t, r):
        b = np.empty(self.n)
        for i in range(1, self.n - 1):
            is_mono = self.m[i - 1] * self.m[i] > 0
            if is_mono:
                b[i] = (
                    3
                    * self.m[i - 1]
                    * self.m[i]
                    / (
                        max(self.m[i - 1], self.m[i])
                        + 2 * min(self.m[i - 1], self.m[i])
                    )
                )
            else:
                b[i] = 0
            if is_mono and self.m[i] > 0:
                b[i] = min(max(0, b[i]), 3 * min(self.m[i - 1], self.m[i]))
            elif is_mono and self.m[i] < 0:
                b[i] = max(min(0, b[i]), 3 * max(self.m[i - 1], self.m[i]))

        b[0] = ((2 * self.h[0] + self.h[1]) * self.m[0] - self.h[0] * self.m[1]) / (
            self.h[0] + self.h[1]
        )
        b[self.n - 1] = (
            (2 * self.h[self.n - 2] + self.h[self.n - 3]) * self.m[self.n - 2]
            - self.h[self.n - 2] * self.m[self.n - 3]
        ) / (self.h[self.n - 2] + self.h[self.n - 3])
        return b

    def evaluate(self, t_intrp):
        ans = []
        for tau in t_intrp:
            i = np.where(tau >= self.x)[0]
            if i.size == 0:
                i = 0
            else:
                i = i[-1]
            i = min(i, self.n - 2)
            res = (
                self.a[i]
                + self.b[i] * (tau - self.x[i])
                + self.c[i] * np.power(tau - self.x[i], 2.0)
                + self.d[i] * np.power(tau - self.x[i], 3.0)
            )  # original curve r(t)
            ans.append(res)
        return ans

    def evaluate_derivative(self, t_intrp):
        ans = []
        if not hasattr(t_intrp, "__len__"):
            t_intrp = np.array([t_intrp])
        for tau in t_intrp:
            i = np.where(tau >= self.x)[0]
            if i.size == 0:
                i = 0
            else:
                i = i[-1]
            i = min(i, self.n - 2)
            res = (
                self.b[i]
                + 2 * self.c[i] * (tau - self.x[i])
                + 3 * self.d[i] * np.power(tau - self.x[i], 2.0)
            )
            ans.append(res)
        if len(ans) == 1:
            return ans[0]
        else:
            return ans

    def evaluate_forward(self, t_intrp):
        ans = []
        for tau in t_intrp:
            i = np.where(tau >= self.x)[0]
            if i.size == 0:
                i = 0
            else:
                i = i[-1]
            i = min(i, self.n - 2)
            res = (
                self.a[i]
                + self.b[i] * (2 * tau - self.x[i])
                + self.c[i] * (tau - self.x[i]) * (3 * tau - self.x[i])
                + self.d[i] * np.power(tau - self.x[i], 2.0) * (4 * tau - self.x[i])
            )  # d(xy)/dx
            ans.append(res)
        return ans



import asyncio
import warnings
from datetime import datetime
from functools import reduce
from typing import Dict, List, Literal, Optional, Tuple

import httpx
import pandas as pd

from CurvyCUSIPs.DataFetcher.base import DataFetcherBase

warnings.filterwarnings("ignore", category=pd.errors.SettingWithCopyWarning)
warnings.simplefilter(action="ignore", category=FutureWarning)

import sys

if sys.platform == "win32":
    loop = asyncio.ProactorEventLoop()
    asyncio.set_event_loop(loop)

import asyncio
import http
import os
import time
import webbrowser
from datetime import date, datetime
from typing import Dict, List, Tuple

import aiohttp
import pandas as pd
import requests


def is_downloadable(url):
    h = requests.head(url, allow_redirects=True)
    header = h.headers
    content_type = header.get("content-type")
    if "text" in content_type.lower():
        return False
    if "html" in content_type.lower():
        return False
    return True


def get_yahoofinance_download_auth(path: str, cj: http.cookiejar = None, run_crumb=False) -> Tuple[dict, str] | dict:
    cookie_str = ""
    if cj:
        cookies = {cookie.name: cookie.value for cookie in cj if "yahoo" in cookie.domain}
        cookies["thamba"] = 2
        cookies["gpp"] = "DBAA"
        cookies["gpp_sid"] = "-1"
        cookie_str = "; ".join([f"{key}={value}" for key, value in cookies.items()])

    headers = {
        "authority": "query1.finance.yahoo.com",
        "method": "GET",
        "path": path,
        "scheme": "https",
        "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7",
        "Accept-Encoding": "gzip, deflate, br",
        "Accept-Language": "en-US,en;q=0.9",
        "Cache-Control": "max-age=0",
        "Cookie": cookie_str,
        "Dnt": "1",
        "Sec-Ch-Ua": '"Chromium";v="116", "Not)A;Brand";v="24", "Google Chrome";v="116"',
        "Sec-Ch-Ua-Mobile": "?0",
        "Sec-Ch-Ua-Platform": '"Windows"',
        "Sec-Fetch-Dest": "document",
        "Sec-Fetch-Mode": "navigate",
        "Sec-Fetch-Site": "none",
        "Sec-Fetch-User": "?1",
        "Upgrade-Insecure-Requests": "1",
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/116.0.0.0 Safari/537.36",
    }
    if not cj:
        del headers["Cookie"]

    if run_crumb:
        crumb_url = "https://query2.finance.yahoo.com/v1/test/getcrumb"
        res = requests.get(crumb_url, headers=headers)
        crumb = res.text

        return headers, crumb

    return headers


def download_historical_data_yahoofinance(ticker: str, from_date: datetime, to_date: datetime, raw_path: str = None, ny_time=False):
    from_sec = round(from_date.timestamp())
    to_sec = round(to_date.timestamp())
    base_url = f"https://query2.finance.yahoo.com/v8/finance/chart/{ticker}?period1={from_sec}&period2={to_sec}&interval=1d&events=history&includeAdjustedClose=true"

    _, crumb = get_yahoofinance_download_auth("/v1/test/getcrumb", None, True)
    headers = get_yahoofinance_download_auth(
        f"/v8/finance/chart/{ticker}?period1={from_sec}&amp;period2={to_sec}&amp;interval=1d&amp;events=history&amp;includeAdjustedClose=true&crumb={crumb}&formatted=false&region=US&lang=en-US",
    )

    res_url = f"{base_url}&crumb={crumb}&formatted=false&region=US&lang=en-US"
    response = requests.get(res_url, headers=headers, allow_redirects=True)
    if response.ok:
        json_data = response.json()
        df = pd.DataFrame(
            {
                "Date": json_data["chart"]["result"][0]["timestamp"],
                "Open": json_data["chart"]["result"][0]["indicators"]["quote"][0]["open"],
                "High": json_data["chart"]["result"][0]["indicators"]["quote"][0]["high"],
                "Low": json_data["chart"]["result"][0]["indicators"]["quote"][0]["low"],
                "Close": json_data["chart"]["result"][0]["indicators"]["quote"][0]["close"],
                "Adj Close": json_data["chart"]["result"][0]["indicators"]["adjclose"][0]["adjclose"],
                "Volume": json_data["chart"]["result"][0]["indicators"]["quote"][0]["volume"],
            }
        )
        df["Date"] = pd.to_datetime(df["Date"], utc=True, unit="s")
        if ny_time:
            df["Date"] = df["Date"].dt.tz_convert("America/New_York")
        else:
            df["Date"] = df["Date"].dt.date

        if raw_path:
            df.to_excel(raw_path)
        return df

    print(f"yf fetch failed with status: {response.status_code}")
    return pd.DataFrame()


def multi_download_historical_data_yahoofinance(
    tickers: List[str],
    from_date: datetime,
    to_date: datetime,
    data_dump_dir: str = None,
    max_date=False,
    big_wb=False,
) -> Dict[str, pd.DataFrame]:
    from_sec = round(from_date.timestamp())
    to_sec = round(to_date.timestamp()) if not max_date else round(datetime.today().timestamp())

    async def fetch(session: aiohttp.ClientSession, url: str, curr_ticker: str, crumb: str) -> pd.DataFrame:
        try:
            headers = get_yahoofinance_download_auth(
                f"/v8/finance/chart/{curr_ticker}?period1={from_sec}&amp;period2={to_sec}&amp;interval=1d&amp;events=history&amp;includeAdjustedClose=true&crumb={crumb}&formatted=false&region=US&lang=en-US",
            )
            res_url = f"{url}&crumb={crumb}&formatted=false&region=US&lang=en-US"
            async with session.get(res_url, headers=headers) as response:
                if response.ok:
                    json_data = await response.json()
                    df = pd.DataFrame(
                        {
                            "Date": json_data["chart"]["result"][0]["timestamp"],
                            "Open": json_data["chart"]["result"][0]["indicators"]["quote"][0]["open"],
                            "High": json_data["chart"]["result"][0]["indicators"]["quote"][0]["high"],
                            "Low": json_data["chart"]["result"][0]["indicators"]["quote"][0]["low"],
                            "Close": json_data["chart"]["result"][0]["indicators"]["quote"][0]["close"],
                            "Adj Close": json_data["chart"]["result"][0]["indicators"]["adjclose"][0]["adjclose"],
                            "Volume": json_data["chart"]["result"][0]["indicators"]["quote"][0]["volume"],
                        }
                    )
                    df["Date"] = pd.to_datetime(df["Date"], utc=True, unit="s")
                    df["Date"] = df["Date"].dt.tz_convert("America/New_York")
                    df["Date"] = df["Date"].dt.tz_localize(None)
                    df["Date"] = df["Date"].dt.date
                    if data_dump_dir:
                        df.to_excel(rf"{data_dump_dir}\{curr_ticker}.xlsx", index=False)
                    return df
                else:
                    raise Exception(f"Bad Status: {response.status} - on {curr_ticker}")
        except Exception as e:
            print(e)
            return pd.DataFrame(columns=["Open", "High", "Low", "Close", "Adj Close", "Volume"])

    async def get_promises(session: aiohttp.ClientSession):
        tasks = []
        _, crumb = get_yahoofinance_download_auth("/v1/test/getcrumb", None, True)
        for ticker in tickers:
            if max_date:
                try:
                    headers = get_yahoofinance_download_auth(
                        f"v8/finance/chart/{ticker}?formatted=true&crumb={crumb}&lang=en-US&region=US&includeAdjustedClose=true&corsDomain=finance.yahoo.com",
                    )
                    first_trade_date_url = f"https://query2.finance.yahoo.com/v8/finance/chart/{ticker}?formatted=true&crumb={crumb}&lang=en-US&region=US&includeAdjustedClose=true&corsDomain=finance.yahoo.com"
                    first_trade_date = requests.get(first_trade_date_url, headers=headers).json()["chart"]["result"][0]["meta"]["firstTradeDate"]
                    from_sec = round(first_trade_date)
                except Exception as e:
                    # print("First Trade Date Error", e)
                    from_sec = round(from_date.timestamp())
            else:
                from_sec = round(from_date.timestamp())

            curr_url = f"https://query2.finance.yahoo.com/v8/finance/chart/{ticker}?period1={from_sec}&period2={to_sec}&interval=1d&events=history&includeAdjustedClose=true"
            task = fetch(session, curr_url, ticker, crumb)
            tasks.append(task)

        return await asyncio.gather(*tasks)

    async def run_fetch_all() -> List[pd.DataFrame]:
        async with aiohttp.ClientSession() as session:
            all_data = await get_promises(session)
            return all_data

    dfs = asyncio.run(run_fetch_all())

    if big_wb:
        tickers_str = str.join("_", [str(x) for x in tickers])
        wb_file_name = f"{data_dump_dir}/{tickers_str}_yahoofin_historical_data.xlsx"
        with pd.ExcelWriter(wb_file_name) as writer:
            for i, df in enumerate(dfs, 0):
                try:
                    df.drop("Unnamed: 0", axis=1, inplace=True)
                except:
                    pass
                df.to_excel(writer, sheet_name=f"{tickers[i]}", index=False)

    return dict(zip(tickers, dfs))


def get_yahoofinance_data_file_path_by_ticker(ticker: str, cj: http.cookiejar = None):
    dir = f"{os.path.abspath('')}/yahoofinance"
    files = sorted(
        os.listdir(dir),
    )
    tickers_with_data = [x for x in files if x.split("_")[0].lower() == ticker.lower()]

    if len(tickers_with_data) == 0:
        from_date = datetime.datetime(2023, 1, 1)
        to_date = datetime.datetime.today()
        download_historical_data_yahoofinance(ticker, from_date, to_date, dir, cj)

    return f"{dir}/{ticker}_yahoofin_historical_data.xlsx"


def get_option_expiration_dates_yahoofinance(ticker: str, cj: http.cookiejar = None) -> List[int]:
    _, crumb = get_yahoofinance_download_auth("/v1/test/getcrumb", cj, True)
    headers = get_yahoofinance_download_auth(
        f"/v7/finance/options/{ticker}?formatted=true&crumb={crumb}&lang=en-US&region=US&date=1&straddle=true&corsDomain=finance.yahoo.com",
        cj,
    )
    url = f"https://query2.finance.yahoo.com/v7/finance/options/{ticker}?formatted=true&crumb=wZdpBzDeWLv&lang=en-US&region=US&date=1&straddle=true&corsDomain=finance.yahoo.com"

    try:
        res = requests.get(url, headers=headers)
        json = res.json()
        return json["optionChain"]["result"][0]["expirationDates"]
    except Exception as e:
        print(e)
        return []


def safe_get(dct, *keys):
    for key in keys:
        try:
            dct = dct[key]
        except KeyError:
            return None
    return dct


def empty_option_data_yahoofinance() -> dict[str, None]:
    return {
        "percentChange": None,
        "openInterest": None,
        "strike": None,
        "change": None,
        "inTheMoney": None,
        "impliedVolatility": None,
        "volume": None,
        "ask": None,
        "contractSymbol": None,
        "lastTradeDate": None,
        "expiration": None,
        "currency": None,
        "contractSize": None,
        "bid": None,
        "lastPrice": None,
    }


def get_options_chain_yahoofinance(
    ticker: str,
    raw_path: str,
    cj: http.cookiejar = None,
    big_wb=False,
) -> Dict[date, pd.DataFrame]:
    async def fetch(session: aiohttp.ClientSession, url: str, ex_date: int, crumb: str) -> pd.DataFrame:
        try:
            headers = get_yahoofinance_download_auth(
                f"/v7/finance/options/{ticker}?formatted=true&crumb={crumb}&lang=en-US&region=US&date={ex_date}&straddle=true&corsDomain=finance.yahoo.com",
                cj,
            )
            async with session.get(url, headers=headers) as response:
                if response.status == 200:
                    # straddle schema
                    """
                    type Straddle = {
                        stike: Option
                    }
                    type Option = {
                        call: OptionInfo
                        puts OptionInfo
                    }
                    type OptionInfo = {
                        "openInterest": int
                        "strike": int
                        "change": int
                        "inTheMoney": bool
                        "impliedVolatility": int
                        "volume": int
                        "ask": int
                        "contractSymbol": str
                        "lastTradeDate": int
                        "expiration": int
                        "currency": int
                        "contractSize": int
                        "bid": int
                        "lastPrice": int
                    }
                    """
                    json = await response.json()
                    data = json["optionChain"]["result"][0]["options"][0]["straddles"]

                    straddle = {}
                    for option in data:
                        strike_price = option["strike"]["raw"]

                        call_raw = safe_get(option, "call")
                        call_raw_dict = call_raw.items() if call_raw else None
                        call = (
                            {key: (value.get("raw", value) if isinstance(value, dict) else value) for key, value in call_raw_dict if call_raw_dict}
                            if call_raw_dict
                            else empty_option_data_yahoofinance()
                        )

                        put_raw = safe_get(option, "put")
                        put_raw_dict = put_raw.items() if put_raw else None
                        put = (
                            {key: (value.get("raw", value) if isinstance(value, dict) else value) for key, value in put_raw_dict if put_raw_dict}
                            if put_raw_dict
                            else empty_option_data_yahoofinance()
                        )

                        straddle[strike_price] = {
                            "call": call,
                            "put": put,
                        }

                    return straddle, ex_date

                else:
                    raise Exception(f"Bad Status: {response.status}")

        except Exception as e:
            print(e)
            return {}

    async def get_promises(session: aiohttp.ClientSession):
        # Option Chain Schema
        """
        type OptionChain = {
            date (int epoch): Straddle
        }
        """
        _, crumb = get_yahoofinance_download_auth("/v1/test/getcrumb", cj, True)
        exp_dates = get_option_expiration_dates_yahoofinance(ticker, cj)
        tasks = []
        for exp_date in exp_dates:
            curr_url = f"https://query2.finance.yahoo.com/v7/finance/options/{ticker}?formatted=true&crumb={crumb}&lang=en-US&region=US&date={exp_date}&straddle=true&corsDomain=finance.yahoo.com"
            task = fetch(session, curr_url, exp_date, crumb)
            tasks.append(task)

        return await asyncio.gather(*tasks)

    """
    return List[Tuple(OptionChain, ExpDate)]
    """

    async def run_fetch_all():
        async with aiohttp.ClientSession() as session:
            all_data = await get_promises(session)
            return all_data

    results = asyncio.run(run_fetch_all())
    dfs = {}
    for_big_wb = []

    try:
        for options_data, exp_date in results:
            strike_prices = list(options_data.keys())
            temp = []
            for strike in strike_prices:
                call_raw_dict = options_data[strike]["call"]
                put_raw_dict = options_data[strike]["put"]

                call_dict = {f"{key}_call": value for key, value in call_raw_dict.items() if key != "strike"}
                put_dict = {f"{key}_put": value for key, value in put_raw_dict.items() if key != "strike"}
                curr_strike_straddle_dict = {**call_dict, **put_dict}
                curr_strike_straddle_dict["strike"] = call_raw_dict["strike"] if call_raw_dict["strike"] else put_raw_dict["strike"]

                print(curr_strike_straddle_dict)
                temp.append(curr_strike_straddle_dict)

                if big_wb:
                    copy_curr_strike_straddle_dict = curr_strike_straddle_dict.copy()
                    copy_curr_strike_straddle_dict["expirationDate"] = exp_date
                    for_big_wb.append(copy_curr_strike_straddle_dict)

            curr_exp_straddle_df = pd.DataFrame(temp)
            curr_exp_straddle_df.set_index("strike")
            dfs[exp_date] = curr_exp_straddle_df

        with pd.ExcelWriter(raw_path, engine="openpyxl") as writer:
            pd.DataFrame().to_excel(writer, index=False)
            for exp_date, df in dfs.items():
                cols = list(df.columns)
                last_price_index = cols.index("lastPrice_call")
                cols.remove("strike")
                cols.insert(last_price_index + 1, "strike")
                df = df[cols]
                df.to_excel(
                    writer,
                    sheet_name=f"{datetime.fromtimestamp(exp_date).strftime('%m-%d-%Y')}",
                    index=False,
                )

            if big_wb:
                big_df = pd.DataFrame(for_big_wb)
                big_df.to_excel(writer, sheet_name="all", index=False)

    except Exception as e:
        print(e)

    return dfs


class YahooFinanceDataFetcher(DataFetcherBase):
    def __init__(
        self,
        global_timeout: int = 10,
        proxies: Optional[Dict[str, str]] = None,
        debug_verbose: Optional[bool] = False,
        info_verbose: Optional[bool] = False,
        error_verbose: Optional[bool] = False,
    ):
        super().__init__(
            global_timeout=global_timeout,
            proxies=proxies,
            debug_verbose=debug_verbose,
            info_verbose=info_verbose,
            error_verbose=error_verbose,
        )

    def fetch_yf_legacy_multi_tickers(
        self,
        tickers: List[str],
        from_date: datetime,
        to_date: datetime,
        data_dump_dir: Optional[str] = None,
        max_date=False,
        big_wb=False,
    ):
        return multi_download_historical_data_yahoofinance(
            tickers=tickers,
            from_date=from_date,
            to_date=to_date,
            data_dump_dir=data_dump_dir,
            max_date=max_date,
            big_wb=big_wb,
        )

    def fetch_yf_legacy(self, ticker: str, from_date: datetime, to_date: datetime, raw_path: Optional[str] = None, ny_time=False):
        return download_historical_data_yahoofinance(ticker=ticker, from_date=from_date, to_date=to_date, raw_path=raw_path, ny_time=ny_time)

    async def _fetch_cusip_timeseries_yahoofinance(
        self,
        client: httpx.AsyncClient,
        cusip: str,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        interval: Optional[Literal["1m", "5m", "1h", "1d"]] = "1h",
        exchange: Optional[Literal["SG", "DU", "MU", "TI"]] = "SG",
        max_retries: Optional[int] = 3,
        backoff_factor: Optional[int] = 1,
        uid: Optional[str | int] = None,
    ):
        pass


# start_date = datetime(2024, 5, 1)
# end_date = datetime(2024, 5, 31)
# period1 = int(start_date.timestamp())
# period2 = int(end_date.timestamp())
# isin = "US912810TN81"
# interval = "1h"
# exchange = "SG"
# # SG, DU, MU, TI

# url = f"https://query1.finance.yahoo.com/v8/finance/chart/{isin}.{exchange}?period1={period1}&period2={period2}&interval={interval}"
# print(url)
# headers = {
#     "authority": "query1.finance.yahoo.com",
#     "method": "GET",
#     "path": url.split(".com")[1],
#     "scheme": "https",
#     "accept": "*/*",
#     "accept-encoding": "gzip, deflate, br, zstd",
#     "accept-language": "en-US,en;q=0.9",
#     "cache-control": "no-cache",
#     "dnt": "1",
#     "origin": "https://finance.yahoo.com",
#     "pragma": "no-cache",
#     "priority": "u=1, i",
#     "referer": f"https://finance.yahoo.com/chart/{isin}.{exchange}",
#     "sec-ch-ua": '"Chromium";v="128", "Not;A=Brand";v="24", "Google Chrome";v="128"',
#     "sec-ch-ua-mobile": "?0",
#     "sec-ch-ua-platform": '"Windows"',
#     "sec-fetch-dest": "empty",
#     "sec-fetch-mode": "cors",
#     "sec-fetch-site": "same-site",
#     "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/127.0.0.0 Safari/537.36",
#     "cookie": r"tbla_id=ce485000-d594-430a-bb0d-106789c83afc-tuctbe93c93; GUC=AQEBCAFm2Z5nAEImMwVs&s=AQAAAAnXOqqR&g=ZthS2Q; A1=d=AQABBA8r7GQCEDl4-HKj5TViLm4-h2a8b8kFEgEBCAGe2WYAZ9w00iMA_eMBAAcIDyvsZGa8b8k&S=AQAAAn93mq3xN8wdsunKFpQ5nYM; A3=d=AQABBA8r7GQCEDl4-HKj5TViLm4-h2a8b8kFEgEBCAGe2WYAZ9w00iMA_eMBAAcIDyvsZGa8b8k&S=AQAAAn93mq3xN8wdsunKFpQ5nYM; A1S=d=AQABBA8r7GQCEDl4-HKj5TViLm4-h2a8b8kFEgEBCAGe2WYAZ9w00iMA_eMBAAcIDyvsZGa8b8k&S=AQAAAn93mq3xN8wdsunKFpQ5nYM; cmp=t=1725546450&j=0&u=1YNN; gpp=DBAA; gpp_sid=-1; PRF=t%3DUS912810UD80.SG%252BUS912810UD8.MU%252BUS912810SX72.SG%252BUS912810UA42.SG%252BUS912810UA4.DU%252B0981.HK%252BUS91282CFG1.BE%252BAMD%252BNVDA%252BJAAA%252BAIR.PA%252BGOVZ%252BIBIT%252BSPY%252BZQQ25.CBT%26newChartbetateaser%3D1%26qke-neo%3Dtrue; axids=gam=y-3cws5sFE2uJzBohwqJrpVs8yxO8XqjLZ~A&dv360=eS1jR0dpUEg1RTJ1RjNjaHF5dnJrckxVVjVhdWJLbG9YY35B&ydsp=y-JF0nspRE2uIqqGlhdNosoSr9syGScSMb~A&tbla=y-gyLskotE2uIge1eqJgtYnhOxvNbazrR3~A",
# }

# res = requests.get(url, headers=headers)
# if res.ok:
#     data = res.json()["chart"]["result"][0]
#     df = pd.DataFrame(
#         {
#             "timestamp": data["timestamp"],
#             "high": data["indicators"]["quote"][0]["high"],
#             "low": data["indicators"]["quote"][0]["low"],
#             "open": data["indicators"]["quote"][0]["open"],
#             "close": data["indicators"]["quote"][0]["close"],
#             "volume": data["indicators"]["quote"][0]["volume"],
#         }
#     )
#     df["timestamp"] = pd.to_datetime(df["timestamp"], utc=True, unit="s")
#     df["timestamp"] = df["timestamp"].dt.tz_convert("America/New_York")

# else:
#     print("rip: ", res.status_code)
#     print(res.content)



from datetime import datetime
from typing import Annotated, Callable, List, Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objs as go
import seaborn as sns
from plotly.subplots import make_subplots
from scipy.interpolate import UnivariateSpline
from scipy.optimize import newton

sns.set_style("whitegrid", {"grid.linestyle": "--"})

import warnings

warnings.simplefilter(action="ignore", category=FutureWarning)


def plot_timeseries(
    df: pd.DataFrame,
    y_cols: List[str],
    x_col="Date",
    max_ticks=10,
    flip=False,
    custom_label_x=None,
    custom_label_y=None,
    custom_title=None,
    date_subset_range: Annotated[List[datetime], 2] | None = None,
    plot_recessions=False,
    bar_plot=False,
    dt_range_highlights: List[Tuple[datetime, datetime, str, str]] = None,
    ohlc=False,
    secondary_y_cols: Optional[List[str]] = None,
    html_path: Optional[str] = None,
):
    copy_df = df.copy()
    date_col = "Date"

    copy_df[date_col] = pd.to_datetime(copy_df[date_col])
    if date_subset_range:
        copy_df = copy_df[(copy_df[date_col] >= date_subset_range[0]) & (copy_df[date_col] <= date_subset_range[1])]
    if flip:
        copy_df = copy_df.iloc[::-1]

    if secondary_y_cols:
        fig = make_subplots(specs=[[{"secondary_y": True}]])
    else:
        fig = go.Figure()

    colors = ["white"] + px.colors.qualitative.Plotly
    for i, y_col in enumerate(y_cols):
        if bar_plot:
            fig.add_trace(
                go.Bar(x=copy_df[x_col], y=copy_df[y_col], name=y_col, marker_color="black"),
                secondary_y=y_col in secondary_y_cols if secondary_y_cols else None,
            )
        elif ohlc:
            fig.add_trace(
                go.Ohlc(
                    x=df[f"Date"],
                    open=df[f"{y_col}_Open"],
                    high=df[f"{y_col}_High"],
                    low=df[f"{y_col}_Low"],
                    close=df[f"{y_col}_Close"],
                    name=y_col,
                    increasing_line_color=colors[i],
                    decreasing_line_color=colors[i],
                ),
                secondary_y=y_col in secondary_y_cols if secondary_y_cols else None,
            )
            fig.update(layout_xaxis_rangeslider_visible=False)
        else:
            fig.add_trace(
                go.Scatter(x=copy_df[x_col], y=copy_df[y_col], mode="lines", name=y_col),
                secondary_y=y_col in secondary_y_cols if secondary_y_cols else None,
            )

    if plot_recessions:
        recessions = [
            [datetime(1961, 4, 1), datetime(1961, 2, 1)],
            [datetime(1969, 12, 1), datetime(1970, 11, 1)],
            [datetime(1973, 11, 1), datetime(1975, 3, 1)],
            [datetime(1980, 1, 1), datetime(1980, 7, 1)],
            [datetime(1981, 7, 1), datetime(1982, 11, 1)],
            [datetime(1990, 7, 1), datetime(1991, 3, 1)],
            [datetime(2001, 3, 1), datetime(2001, 11, 1)],
            [datetime(2007, 12, 1), datetime(2009, 6, 1)],
            [datetime(2020, 2, 1), datetime(2020, 4, 1)],
        ]
        if date_subset_range:
            start_plot_range, end_plot_range = min(date_subset_range), max(date_subset_range)
        else:
            start_plot_range, end_plot_range = min(copy_df["Date"]), max(copy_df["Date"])

        for recession_dates in recessions:
            start_date, end_date = recession_dates
            if start_date <= end_plot_range and end_date >= start_plot_range:
                fig.add_vrect(
                    x0=start_date,
                    x1=end_date,
                    fillcolor="red",
                    opacity=0.3,
                    layer="below",
                    line_width=0,
                )

    if dt_range_highlights:
        for highlight_props in dt_range_highlights:
            start_date, end_date, color, title = highlight_props
            fig.add_vrect(
                x0=start_date,
                x1=end_date,
                fillcolor=color,
                opacity=0.3,
                layer="below",
                line_width=0,
                annotation_text=title,
                annotation_position="top left",
                annotation_textangle=90,
            )

    if secondary_y_cols:
        fig.update_yaxes(title=custom_label_y or ", ".join(secondary_y_cols), secondary_y=True)
        fig.update_yaxes(title=custom_label_y or ", ".join(list(set(y_cols) - set(secondary_y_cols))), secondary_y=False)
    else:
        fig.update_yaxes(title=custom_label_y or ", ".join(y_cols))
    fig.update_layout(
        xaxis_title=custom_label_x or x_col,
        xaxis=dict(nticks=max_ticks),
        title=custom_title or "Yield Plot",
        showlegend=True,
        template="plotly_dark",
        height=700,
    )
    fig.update_xaxes(showspikes=True, spikecolor="white", spikesnap="cursor", spikemode="across")
    fig.update_yaxes(showspikes=True, spikecolor="white", spikesnap="cursor", spikethickness=0.5)
    fig.show(
        config={
            "modeBarButtonsToAdd": [
                "drawline",
                "drawopenpath",
                "drawclosedpath",
                "drawcircle",
                "drawrect",
                "eraseshape",
            ]
        }
    )
    if html_path:
        fig.write_html(html_path)


def plot_yield_curve_date_range(
    df: pd.DataFrame,
    date_range: List[datetime],
    reverse_mats=False,
    real_yields=False,
    adjusted=False,
    del_mats=None,
):
    maturities = df.columns[1:]
    months = [6, 12, 24, 36, 60, 84, 120, 240, 360]
    months = [50, 84, 120, 240, 360] if real_yields else months
    if adjusted:
        months = [0, 10, 20, 30, 60, 90, 110, 130, 170, 200, 280, 370, 520]
        months = [0, 10, 25, 50, 75] if real_yields else months

    if del_mats:
        df = df.drop(columns=del_mats)
        indexes = [maturities.index(to_del) for to_del in del_mats]
        maturities = [i for j, i in enumerate(maturities) if j not in indexes]
        months = [i for j, i in enumerate(months) if j not in indexes]

    fig = go.Figure()
    for date in date_range:
        try:
            formatted_date = date.strftime("%Y-%m-%d")
            tooltip_formatted_date = date.strftime("%m/%d/%Y")
            yields = df.loc[df["Date"] == formatted_date]
            yields = yields.values.tolist()[0]
            del yields[0]

            if reverse_mats:
                yields.reverse()
                maturities.reverse()

            fig.add_trace(
                go.Scatter(
                    x=months,
                    y=yields,
                    mode="lines+markers",
                    name=f"{tooltip_formatted_date}",
                )
            )
        except:
            print(f"Failed to plot: {date}")

    formatted_title = ""
    for dr in date_range:
        formatted_title += f"{dr.strftime('%Y-%m-%d')} vs "
    formatted_title = formatted_title[:-3]

    fig.update_layout(
        title=formatted_title if not real_yields else f"{formatted_title} (Real)",
        xaxis_title="Maturity (Months)",
        yaxis_title="Yield (%)",
        xaxis=dict(tickvals=months, ticktext=maturities),
        showlegend=True,
        height=700,
        template="plotly_dark",
    )
    fig.update_xaxes(showspikes=True, spikecolor="white", spikesnap="cursor", spikemode="across")
    fig.update_yaxes(showspikes=True, spikecolor="white", spikesnap="cursor", spikethickness=0.5)
    fig.show(
        config={
            "modeBarButtonsToAdd": [
                "drawline",
                "drawopenpath",
                "drawclosedpath",
                "drawcircle",
                "drawrect",
                "eraseshape",
            ]
        }
    )


def par_curve_func(tenor, zero_curve_func, is_parametric_class=False):
    if is_parametric_class:

        def par_bond_equation(c, maturity, zero_curve_func):
            discounted_cash_flows = sum((c / 2) * np.exp(-(zero_curve_func([t])[0] / 100) * t) for t in np.arange(0.5, maturity + 0.5, 0.5))
            final_payment = 100 * np.exp(-(zero_curve_func([maturity])[0] / 100) * maturity)
            return discounted_cash_flows + final_payment - 100

    else:

        def par_bond_equation(c, maturity, zero_curve_func):
            discounted_cash_flows = sum((c / 2) * np.exp(-(zero_curve_func(t) / 100) * t) for t in np.arange(0.5, maturity + 0.5, 0.5))
            final_payment = 100 * np.exp(-(zero_curve_func(maturity) / 100) * maturity)
            return discounted_cash_flows + final_payment - 100

    init_guess = 4
    return newton(
        par_bond_equation,
        x0=init_guess,
        args=(tenor, zero_curve_func),
    )


def plot_usts(
    curve_set_df: pd.DataFrame,
    ttm_col: Optional[str] = "time_to_maturity",
    ytm_col: Optional[str] = "ytm",
    label_col: Optional[str] = "original_security_term",
    hover_data: Optional[List[str]] = None,
    title: Optional[str] = None,
    custom_x_axis: Optional[str] = "Time to Maturity",
    custom_y_axis: Optional[str] = "Yield to Maturity",
    zero_curves: Optional[List[Tuple[Callable, str]]] = None,
    par_curves: Optional[List[Tuple[Callable, str, bool]]] = None,
    n_yr_fwd_curves: Optional[List[Tuple[Callable, int, str]]] = None,
    impl_spot_n_yr_fwd_curves: Optional[List[Tuple[Callable, int, str]]] = None,
    impl_par_n_yr_fwd_curves: Optional[List[Tuple[Callable, int, str, bool]]] = None,
    cusips_filter: Optional[List[str]] = None,
    ust_labels_filter: Optional[List[str]] = None,
    cusips_hightlighter: Optional[List[str]] = None,
    ust_labels_highlighter: Optional[List[Tuple[str, str] | str]] = None,
    linspace_num: Optional[int] = 1000,
    y_axis_range: Optional[Annotated[List[int], 2]] = [3.33, 5.5],
    plot_height=1000,
    plot_width=None,
):
    curve_set_df = curve_set_df.copy()

    if cusips_filter:
        curve_set_df = curve_set_df[curve_set_df["cusip"].isin(cusips_filter)]
        cusips_filter_set = set(cusips_filter)
        cusips_in_df = set(curve_set_df["cusip"].unique())
        cusips_not_in_df = cusips_filter_set - cusips_in_df
        print("CUSIPs not in Curveset df:", cusips_not_in_df)

    curve_set_df["int_rate"] = pd.to_numeric(curve_set_df["int_rate"], errors="coerce")
    curve_set_df["ust_label"] = curve_set_df.apply(
        lambda row: f"{row['int_rate']:.3f}% {row['maturity_date'].strftime('%b-%y')}",
        axis=1,
    )
    if ust_labels_filter:
        curve_set_df = curve_set_df[curve_set_df["ust_label"].isin(ust_labels_filter)]
        ust_labels_filter_set = set(ust_labels_filter)
        labels_in_df = set(curve_set_df["ust_label"].unique())
        labels_not_in_df = ust_labels_filter_set - labels_in_df
        print("Labels not in Curveset df:", labels_not_in_df)

    otr_mask = curve_set_df["rank"] == 0
    fig = px.scatter(
        curve_set_df[~otr_mask],
        x=ttm_col,
        y=ytm_col,
        color=label_col,
        hover_data=hover_data,
    )

    curve_set_df.loc[otr_mask, label_col] = curve_set_df.loc[otr_mask, label_col].apply(lambda x: f"OTR - {x}")
    otr_fig = px.scatter(
        curve_set_df[otr_mask],
        x=ttm_col,
        y=ytm_col,
        color=label_col,
        hover_data=hover_data,
    )
    for trace in otr_fig.data:
        trace.update(
            marker=dict(
                line=dict(color="white", width=2),
            )
        )
    fig.add_traces(otr_fig.data)

    if cusips_hightlighter:
        if isinstance(cusips_hightlighter[0], tuple):
            for cusip_tuple in cusips_hightlighter:
                if not isinstance(cusip_tuple, tuple):
                    cusip = cusip_tuple
                    label_color = "yellow"
                else:
                    cusip, label_color = cusip_tuple

                if cusip not in curve_set_df["cusip"].values:
                    print(f"{cusip} not in Curveset df!")
                    continue

                cusip_highlight_fig = px.scatter(
                    curve_set_df[curve_set_df["cusip"] == cusip],
                    x=ttm_col,
                    y=ytm_col,
                    color="cusip",
                    hover_data=hover_data,
                )
                for trace in cusip_highlight_fig.data:
                    trace.update(
                        marker=dict(
                            line=dict(color=label_color, width=4),
                        )
                    )
                fig.add_traces(cusip_highlight_fig.data)
        else:
            cusip_highlight_mask = curve_set_df["cusip"].isin(cusips_hightlighter)
            cusip_highlight_fig = px.scatter(
                curve_set_df[cusip_highlight_mask],
                x=ttm_col,
                y=ytm_col,
                color="cusip",
                hover_data=hover_data,
            )
            for trace in cusip_highlight_fig.data:
                trace.update(
                    marker=dict(
                        line=dict(color="yellow", width=4),
                    )
                )
            fig.add_traces(cusip_highlight_fig.data)

    if ust_labels_highlighter:
        if isinstance(ust_labels_highlighter[0], tuple):
            for label_tuple in ust_labels_highlighter:
                if not isinstance(label_tuple, tuple):
                    ust_label = label_tuple
                    label_color = "yellow"
                else:
                    ust_label, label_color = label_tuple

                if ust_label not in curve_set_df["ust_label"].values:
                    print(f"{ust_label} not in Curveset df!")
                    continue

                ust_labels_highlight_fig = px.scatter(
                    curve_set_df[curve_set_df["ust_label"] == ust_label],
                    x=ttm_col,
                    y=ytm_col,
                    color="ust_label",
                    hover_data=hover_data,
                )
                for trace in ust_labels_highlight_fig.data:
                    trace.update(
                        marker=dict(
                            line=dict(color=label_color, width=4),
                        )
                    )
                fig.add_traces(ust_labels_highlight_fig.data)
        else:
            ust_labels_highlight_mask = curve_set_df["ust_label"].isin(ust_labels_highlighter)
            ust_labels_highlight_fig = px.scatter(
                curve_set_df[ust_labels_highlight_mask],
                x=ttm_col,
                y=ytm_col,
                color="ust_label",
                hover_data=hover_data,
            )
            for trace in ust_labels_highlight_fig.data:
                trace.update(
                    marker=dict(
                        line=dict(color="yellow", width=4),
                    )
                )
            fig.add_traces(ust_labels_highlight_fig.data)

    if zero_curves:
        ttm_linspace = np.linspace(0, 30, linspace_num)
        for curve_tup in zero_curves:
            interp_func, label = curve_tup
            fig.add_trace(
                go.Scatter(
                    x=ttm_linspace,
                    y=interp_func(ttm_linspace),
                    mode="lines",
                    name=label,
                )
            )

        zero_spline, label = zero_curves[0]
        residuals = curve_set_df[ytm_col].to_numpy() - zero_spline(curve_set_df[ttm_col].to_numpy())
        plt.figure(figsize=(20, 10))
        plt.scatter(curve_set_df[ttm_col], residuals * 100, color="b", label="Residuals (bps)")
        plt.axhline(0, color="r", linestyle="--")
        plt.xlabel(ttm_col)
        plt.ylabel("Residuals (bps)")
        plt.title(f"Residuals of {label} Spline Fit", fontdict={"fontsize": "x-large"})
        plt.legend(fontsize="x-large")
        plt.ylim(-50, 50)

    if par_curves:
        cfs = np.arange(0.5, 30 + 1, 0.5)
        for curve_tup in par_curves:
            if not len(curve_tup) == 3:
                curve_tup = curve_tup + (False,)
            interp_func, label, is_parametric_class = curve_tup
            fig.add_trace(
                go.Scatter(
                    x=cfs,
                    y=[par_curve_func(t, interp_func, is_parametric_class) for t in cfs],
                    mode="lines",
                    name=label,
                )
            )

    if n_yr_fwd_curves:
        ttm_linspace = np.linspace(0.5, 30, linspace_num)
        for curve_tup in n_yr_fwd_curves:
            interp_func, n, label = curve_tup
            Z_t = interp_func(ttm_linspace)

            if n == -1 or n == np.inf:
                try:
                    interp_func_derivative = interp_func.derivative()
                    dZ_dt = interp_func_derivative(ttm_linspace)
                    F_t = Z_t + ttm_linspace * dZ_dt
                    fig.add_trace(
                        go.Scatter(
                            x=ttm_linspace,
                            y=F_t,
                            mode="lines",
                            name=label,
                        )
                    )
                except Exception as e:
                    print("only scipy interpolation function are supported for inst fwds")
                    print(e)
            else:
                forward_rates = []
                for t1 in ttm_linspace:
                    t2 = t1 + n
                    Z_t1 = interp_func(t1)
                    Z_t2 = interp_func(t2)
                    fwd_rate = (Z_t2 * t2 - Z_t1 * t1) / (t2 - t1)
                    forward_rates.append(fwd_rate)

                fig.add_trace(
                    go.Scatter(
                        x=ttm_linspace,
                        y=forward_rates,
                        mode="lines",
                        name=label,
                    )
                )

    if impl_spot_n_yr_fwd_curves:
        cfs = np.arange(0.5, 30 + 1, 0.5)
        for curve_tup in impl_spot_n_yr_fwd_curves:
            interp_func, n, label = curve_tup
            implied_spot_rates = []
            for t in cfs:
                if t > n:
                    Z_t_temp = interp_func(t)
                    Z_n = interp_func(n)
                    Z_n_t = (Z_t_temp * t - Z_n * n) / (t - n)
                    implied_spot_rates.append(Z_n_t)
                else:
                    implied_spot_rates.append(np.nan)
            implied_spot_rates = np.array(implied_spot_rates)
            fig.add_trace(
                go.Scatter(
                    x=cfs,
                    y=implied_spot_rates,
                    mode="lines",
                    name=label,
                )
            )

    if impl_par_n_yr_fwd_curves:
        cfs = np.arange(0.5, 30 + 1, 0.5)
        for curve_tup in impl_par_n_yr_fwd_curves:
            if not len(curve_tup) == 4:
                curve_tup = curve_tup + (False,)

            interp_func, n, label, is_parametric_class = curve_tup
            implied_spot_rates_n_fwd = []
            if is_parametric_class:
                for t in cfs:
                    if t > n:
                        Z_t_temp = interp_func([t])[0]
                        Z_n = interp_func([n])[0]
                        Z_n_t = (Z_t_temp * t - Z_n * n) / (t - n)
                        implied_spot_rates_n_fwd.append(Z_n_t)
                    else:
                        implied_spot_rates_n_fwd.append(np.nan)
            else:
                for t in cfs:
                    if t > n:
                        Z_t_temp = interp_func(t)
                        Z_n = interp_func(n)
                        Z_n_t = (Z_t_temp * t - Z_n * n) / (t - n)
                        implied_spot_rates_n_fwd.append(Z_n_t)
                    else:
                        implied_spot_rates_n_fwd.append(np.nan)
            implied_spot_rates_n_fwd = np.array(implied_spot_rates_n_fwd)[n * 2 :]
            implied_par_rates_n_fwd_spline = UnivariateSpline(cfs[n * 2 :], implied_spot_rates_n_fwd, s=0, k=3)
            fig.add_trace(
                go.Scatter(
                    x=cfs[n * 2 :],
                    y=[par_curve_func(t, implied_par_rates_n_fwd_spline, is_parametric_class) for t in cfs[n * 2 :]],
                    mode="lines",
                    name=label,
                )
            )

    fig.update_layout(
        xaxis_title=custom_x_axis,
        yaxis_title=custom_y_axis,
        title=title or "Yield Curve",
        showlegend=True,
        template="plotly_dark",
        height=plot_height,
        width=plot_width,
    )
    fig.update_xaxes(showspikes=True, spikecolor="white", spikesnap="cursor", spikemode="across")
    fig.update_yaxes(
        showspikes=True,
        spikecolor="white",
        spikesnap="cursor",
        spikethickness=0.5,
        range=y_axis_range,
    )
    fig.show(
        config={
            "modeBarButtonsToAdd": [
                "drawline",
                "drawopenpath",
                "drawclosedpath",
                "drawcircle",
                "drawrect",
                "eraseshape",
            ]
        }
    )
    plt.show()



from typing import List, Optional, Union

import numpy as np
from scipy import optimize
from scipy.optimize import minimize


class SmithWilsonCurve:
    def __init__(self, ufr: float, alpha: Optional[float] = None):
        self.ufr = ufr  # Ultimate Forward Rate (annualized/annual compounding)
        self.alpha = alpha  # Convergence speed parameter
        self.zeta = None  # To be determined after fitting the model
        self.t_obs = None  # Observed maturities
        self.rates_obs = None  # Observed zero-coupon rates

    @staticmethod
    def calculate_prices(
        rates: Union[np.ndarray, List[float]], t: Union[np.ndarray, List[float]]
    ) -> np.ndarray:
        rates = np.array(rates)
        t = np.array(t)
        return np.power(1 + rates, -t)

    def ufr_discount_factor(self, t: Union[np.ndarray, List[float]]) -> np.ndarray:
        ufr_log = np.log(1 + self.ufr)
        t = np.array(t)
        return np.exp(-ufr_log * t)

    def wilson_function(
        self, t1: Union[np.ndarray, List[float]], t2: Union[np.ndarray, List[float]]
    ) -> np.ndarray:
        t1 = np.array(t1).reshape(-1, 1)  # Ensure t1 is a column vector
        t2 = np.array(t2).reshape(-1, 1)  # Ensure t2 is a column vector

        m = len(t1)
        n = len(t2)

        t1_Mat = np.repeat(t1, n, axis=1)
        t2_Mat = np.repeat(t2, m, axis=1).T

        min_t = np.minimum(t1_Mat, t2_Mat)
        max_t = np.maximum(t1_Mat, t2_Mat)

        ufr_disc = self.ufr_discount_factor(t=(t1_Mat + t2_Mat))
        W = ufr_disc * (
            self.alpha * min_t
            - 0.5
            * np.exp(-self.alpha * max_t)
            * (np.exp(self.alpha * min_t) - np.exp(-self.alpha * min_t))
        )
        return W

    def fit_parameters(
        self, rates: Union[np.ndarray, List[float]], t: Union[np.ndarray, List[float]]
    ) -> np.ndarray:
        W = self.wilson_function(t1=t, t2=t)
        mu = self.ufr_discount_factor(t=t)
        P = self.calculate_prices(rates=rates, t=t)

        zeta = np.linalg.inv(W) @ (mu - P)
        self.zeta = zeta
        return zeta

    def fit_smithwilson_rates(
        self,
        rates_obs: Union[np.ndarray, List[float]],
        t_obs: Union[np.ndarray, List[float]],
        t_target: Union[np.ndarray, List[float]],
    ) -> np.ndarray:
        rates_obs = np.array(rates_obs).reshape((-1, 1))
        t_obs = np.array(t_obs).reshape((-1, 1))
        t_target = np.array(t_target).reshape((-1, 1))

        if self.alpha is None:
            self.alpha = self.fit_convergence_parameter(
                rates_obs=rates_obs, t_obs=t_obs
            )

        self.fit_parameters(rates=rates_obs, t=t_obs)
        ufr_disc = self.ufr_discount_factor(t=t_target)
        W = self.wilson_function(t1=t_target, t2=t_obs)

        P = ufr_disc - W @ self.zeta
        return np.power(1 / P, 1 / t_target) - 1

    def fit_convergence_parameter(
        self,
        rates_obs: Union[np.ndarray, List[float]],
        t_obs: Union[np.ndarray, List[float]],
    ) -> float:
        llp = np.max(t_obs)
        convergence_t = max(llp + 40, 60)

        def forward_difference(alpha: float):
            # Set alpha and fit the model
            self.alpha = alpha
            rates = self.fit_smithwilson_rates(
                rates_obs=rates_obs,
                t_obs=t_obs,
                t_target=[convergence_t, convergence_t + 1],
            )

            forward_rate = (1 + rates[1]) ** (convergence_t + 1) / (
                1 + rates[0]
            ) ** convergence_t - 1
            return -abs(forward_rate - self.ufr) + 1 / 10_000

        result = optimize.minimize(
            lambda alpha: alpha,
            x0=0.15,
            method="SLSQP",
            bounds=[[0.05, 1.0]],
            constraints=[{"type": "ineq", "fun": forward_difference}],
            options={"ftol": 1e-6, "disp": True},
        )

        return float(result.x)

    def fit(
        self,
        rates_obs: Union[np.ndarray, List[float]],
        t_obs: Union[np.ndarray, List[float]],
    ):
        self.t_obs = t_obs
        self.rates_obs = rates_obs
        if self.alpha is None:
            self.alpha = self.fit_convergence_parameter(
                rates_obs=rates_obs, t_obs=t_obs
            )
        self.fit_parameters(rates=rates_obs, t=t_obs)

    def interpolate(self, t_target: Union[np.ndarray, List[float]]) -> np.ndarray:
        return self.fit_smithwilson_rates(self.rates_obs, self.t_obs, t_target)

    def __call__(self, t_target: Union[np.ndarray, List[float]]) -> np.ndarray:
        return self.interpolate(t_target)


# https://www.sciencedirect.com/science/article/pii/S0167668723000963
def find_ufr_ytm(
    maturities, ytms, alpha_min=0.05, tau=0.0001, max_iter=100, ufr_guess=0.04
):
    def wilson_function_matrix(alpha, t):
        # Compute Wilson function matrix W_alpha for given alpha and maturities t
        W = np.zeros((len(t), len(t)))
        for i in range(len(t)):
            for j in range(len(t)):
                W[i, j] = (
                    alpha * min(t[i], t[j])
                    - 0.5 * np.exp(-alpha * abs(t[i] - t[j]))
                    + 0.5 * np.exp(-alpha * (t[i] + t[j]))
                )
        return W

    def objective(ufr, alpha):
        W_alpha = wilson_function_matrix(alpha, maturities)
        discount_factors = np.exp(-ytms * np.array(maturities))
        implied_prices = np.exp(-ufr * np.array(maturities))

        # Ensure the smoothness and correct the discount curve using W_alpha
        residuals = discount_factors - implied_prices
        penalty = np.sum((residuals**2) + alpha**2 * np.dot(W_alpha, residuals) ** 2)

        # Additional penalty based on the convergence condition with tau
        forward_rate_convergence = np.abs(
            residuals[-1]
        )  # Assume last residual represents forward rate convergence
        if forward_rate_convergence > tau:
            penalty += (
                1e6 * (forward_rate_convergence - tau) ** 2
            )  # Large penalty if it exceeds tau

        return penalty

    def find_optimal_alpha(ufr):
        result = minimize(
            lambda alpha: objective(ufr, alpha),
            x0=0.1,
            bounds=[(alpha_min, 1.0)],
            options={"maxiter": max_iter},
        )
        return result.x[0]

    alpha = find_optimal_alpha(ufr_guess)

    result = minimize(
        lambda ufr: objective(ufr, alpha),
        x0=ufr_guess,
        bounds=[(0, None)],
        options={"maxiter": max_iter},
    )
    ufr_optimal = result.x[0]

    return ufr_optimal



import asyncio
import warnings
from datetime import datetime
from functools import reduce
from typing import Dict, List, Optional, Tuple

import httpx
import pandas as pd
import ujson as json
from requests.models import PreparedRequest

from CurvyCUSIPs.DataFetcher.base import DataFetcherBase

warnings.simplefilter(action="ignore", category=FutureWarning)

import sys

if sys.platform == "win32":
    loop = asyncio.ProactorEventLoop()
    asyncio.set_event_loop(loop)


class WSJDataFetcher(DataFetcherBase):
    def __init__(
        self,
        global_timeout: int = 10,
        proxies: Optional[Dict[str, str]] = None,
        debug_verbose: Optional[bool] = False,
        info_verbose: Optional[bool] = False,
        error_verbose: Optional[bool] = False,
    ):
        super().__init__(
            global_timeout=global_timeout,
            proxies=proxies,
            debug_verbose=debug_verbose,
            info_verbose=info_verbose,
            error_verbose=error_verbose,
        )

    async def _fetch_timeseries(
        self,
        client: httpx.AsyncClient,
        wsj_ticker_key: str,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        max_retries: Optional[int] = 3,
        backoff_factor: Optional[int] = 1,
        uid: Optional[str | int] = None,
    ):
        payload = {
            "Step": "P1D",
            "TimeFrame": "all",
            "EntitlementToken": "57494d5ed7ad44af85bc59a51dd87c90",
            "IncludeMockTick": True,
            "FilterNullSlots": True,
            "FilterClosedPoints": True,
            "IncludeClosedSlots": True,
            "IncludeOfficialClose": True,
            "InjectOpen": True,
            "ShowPreMarket": True,
            "ShowAfterHours": True,
            "UseExtendedTimeFrame": True,
            "WantPriorClose": True,
            "IncludeCurrentQuotes": True,
            "ResetTodaysAfterHoursPercentChange": False,
            "Series": [
                {
                    "Key": wsj_ticker_key,
                    "Dialect": "Charting",
                    "Kind": "Ticker",
                    "SeriesId": "s1",
                    "DataTypes": ["Last"],
                }
            ],
        }
        params = {
            "json": json.dumps(payload),
            "ckey": "57494d5ed7",
        }
        url = "https://api.wsj.net/api/michelangelo/timeseries/history"
        prep_url = PreparedRequest()
        prep_url.prepare_url(url, params)
        headers = {
            "authority": "api.wsj.net",
            "method": "GET",
            "path": prep_url.path_url,
            "scheme": "https",
            "Connection": "keep-alive",
            "Pragma": "no-cache",
            "Cache-Control": "no-cache",
            "sec-ch-ua": '" Not A;Brand";v="99", "Chromium";v="99", "Google Chrome";v="99"',
            "Accept": "application/json, text/javascript, */*; q=0.01",
            "Dylan2010.EntitlementToken": "57494d5ed7ad44af85bc59a51dd87c90",
            "sec-ch-ua-mobile": "?0",
            "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/99.0.4844.74 Safari/537.36",
            "sec-ch-ua-platform": '"macOS"',
            "Origin": "https://www.wsj.com",
            "Sec-Fetch-Site": "cross-site",
            "Sec-Fetch-Mode": "cors",
            "Sec-Fetch-Dest": "empty",
            "Referer": "https://www.wsj.com/",
            "Accept-Language": "en-US,en;q=0.9,es;q=0.8",
        }

        cols_to_return = ["Date", wsj_ticker_key]
        retries = 0
        try:
            while retries < max_retries:
                try:
                    response = await client.get(prep_url.url, headers=headers)
                    response.raise_for_status()
                    json_data = response.json()
                    df = pd.DataFrame({"Date": json_data["TimeInfo"]["Ticks"], wsj_ticker_key: [d[0] for d in json_data["Series"][0]["DataPoints"]]})
                    df["Date"] = pd.to_datetime(df["Date"], unit="ms")
                    if start_date:
                        df = df[df["Date"] >= start_date]
                    if end_date:
                        df = df[df["Date"] <= end_date]

                    if uid:
                        return wsj_ticker_key, df, uid
                    return wsj_ticker_key, df

                except httpx.HTTPStatusError as e:
                    self._logger.error(f"WSJ - Bad Status: {response.status_code}")
                    if response.status_code == 404:
                        if uid:
                            return wsj_ticker_key, pd.DataFrame(columns=cols_to_return), uid
                        return wsj_ticker_key, pd.DataFrame(columns=cols_to_return)

                    retries += 1
                    wait_time = backoff_factor * (2 ** (retries - 1))
                    self._logger.debug(f"WSJ- Throttled for {wsj_ticker_key}. Waiting for {wait_time} seconds before retrying...")
                    await asyncio.sleep(wait_time)

                except Exception as e:
                    self._logger.error(f"WSJ - Error: {str(e)}")
                    retries += 1
                    wait_time = backoff_factor * (2 ** (retries - 1))
                    self._logger.debug(f"WSJ - Throttled for {wsj_ticker_key}. Waiting for {wait_time} seconds before retrying...")
                    await asyncio.sleep(wait_time)

            raise ValueError(f"WSJ  - Max retries exceeded for {wsj_ticker_key}")

        except Exception as e:
            self._logger.error(e)
            if uid:
                return wsj_ticker_key, pd.DataFrame(columns=cols_to_return), uid
            return wsj_ticker_key, pd.DataFrame(columns=cols_to_return)

    async def _fetch_timeseries_with_semaphore(self, semaphore, *args, **kwargs):
        async with semaphore:
            return await self._fetch_timeseries(*args, **kwargs)

    def wsj_timeseries_api(
        self,
        wsj_ticker_keys: List[str],
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        max_concurrent_tasks: int = 64,
        merge_into_one_df: Optional[bool] = False,
    ):
        async def build_tasks(
            client: httpx.AsyncClient,
            wsj_ticker_keys: List[str],
            start_date: datetime,
            end_date: datetime,
        ):
            semaphore = asyncio.Semaphore(max_concurrent_tasks)
            tasks = [
                self._fetch_timeseries_with_semaphore(
                    semaphore=semaphore,
                    client=client,
                    wsj_ticker_key=wsj_ticker_key,
                    start_date=start_date,
                    end_date=end_date,
                )
                for wsj_ticker_key in wsj_ticker_keys
            ]
            return await asyncio.gather(*tasks)

        async def run_fetch_all(
            wsj_ticker_keys: List[str],
            start_date: datetime,
            end_date: datetime,
        ):
            async with httpx.AsyncClient(proxy=self._proxies["https"]) as client:
                all_data = await build_tasks(
                    client=client,
                    wsj_ticker_keys=wsj_ticker_keys,
                    start_date=start_date,
                    end_date=end_date,
                )
                return all_data

        dfs: List[Tuple[str, pd.DataFrame]] = asyncio.run(
            run_fetch_all(
                wsj_ticker_keys=wsj_ticker_keys,
                start_date=start_date,
                end_date=end_date,
            )
        )

        if merge_into_one_df:
            merge_dfs_on_column = lambda dfs_dict, on_column: reduce(lambda left, right: pd.merge(left, right, on=on_column), dfs_dict.values())
            return merge_dfs_on_column(dict(dfs), "Date")

        return dict(dfs)
